from __future__ import annotations

import argparse
from pathlib import Path
from typing import Literal

import more_itertools as itx
import pytest
from hypothesis import given
from hypothesis import strategies as st

from snakebids.exceptions import ConfigError
from snakebids.plugins.cli_config import CliConfig


class TestMakesUnderscoreDashAliases:
    @given(
        args=st.lists(
            st.from_regex(r"[^-_\n]+-[^-_\n]+_[^-_\n]+", fullmatch=True),
            max_size=5,
            unique=True,
        )
    )
    @pytest.mark.parametrize("hot", [("dashed",), ("scored",)])
    def test_underscores_to_dashes(
        self, args: list[str], hot: Literal["dashed", "scored"]
    ):
        cli = CliConfig()
        parser = argparse.ArgumentParser()
        dashed = ["--" + s.replace("_", "-") for s in args]
        scored = ["--" + s.replace("-", "_") for s in args]
        curr = dashed if hot == "dashed" else scored
        opp = scored if hot == "dashed" else dashed
        config = {"parse_args": {arg: {"action": "store_true"} for arg in curr}}
        cli.add_cli_arguments(parser, config)
        nm = parser.parse_args(opp)
        assert all(nm.__dict__.values())
        assert len(nm.__dict__) == len(args)

    @given(arg=st.text().filter(lambda s: not s.startswith("-") and len(s)))
    def test_positional_args_added_without_conversion(self, arg: str):
        cli = CliConfig()
        parser = argparse.ArgumentParser()
        config = {"parse_args": {arg: {"help": "..."}}}
        cli.add_cli_arguments(parser, config)
        nm = parser.parse_args(["..."])
        assert itx.one(nm.__dict__.values()) == "..."

    def test_fails_if_undefined_type_given(self):
        cli = CliConfig()
        config = {"parse_args": {"--new-param": {"type": "UnheardClass"}}}
        parser = argparse.ArgumentParser()
        with pytest.raises(ConfigError, match="could not be resolved"):
            cli.add_cli_arguments(parser, config)


def test_works_when_key_not_present():
    cli = CliConfig()
    parser = argparse.ArgumentParser()
    cli.add_cli_arguments(parser, {})
    assert len(parser._actions) == 1


def test_convert_arg_to_builtin():
    cli = CliConfig()
    parser = argparse.ArgumentParser()
    new_args = {
        "--new-param": {
            "help": "Generic Help Message",
            "type": "int",
        }
    }
    cli.add_cli_arguments(parser, {"parse_args": new_args})
    args = parser.parse_args(["--new-param", "12"])
    assert isinstance(args.new_param, int)


def test_path_works_as_builtin():
    cli = CliConfig()
    parser = argparse.ArgumentParser()
    new_args = {
        "--new-param": {
            "help": "Generic Help Message",
            "type": "Path",
        }
    }
    cli.add_cli_arguments(parser, {"parse_args": new_args})
    args = parser.parse_args(["--new-param", "foo"])
    assert args.new_param == Path("foo")


def test_non_serialiable_type_raises_error():
    cli = CliConfig()
    parser = argparse.ArgumentParser()
    new_args = {
        "--new-param": {
            "help": "Generic Help Message",
            "type": "snakebids.utils.utils.BidsEntity",
        }
    }
    with pytest.raises(ConfigError, match="cannot be serialized into yaml"):
        cli.add_cli_arguments(parser, {"parse_args": new_args})


def test_using_module_as_type_gives_error():
    cli = CliConfig()
    parser = argparse.ArgumentParser()
    new_args = {
        "--new-param": {
            "help": "Generic Help Message",
            "type": "snakebids.utils.utils",
        }
    }
    with pytest.raises(ConfigError, match="cannot be used as a type"):
        cli.add_cli_arguments(parser, {"parse_args": new_args})


def test_using_class_method_as_type_gives_error():
    cli = CliConfig()
    parser = argparse.ArgumentParser()
    new_args = {
        "--new-param": {
            "help": "Generic Help Message",
            "type": "snakebids.utils.utils.BidsEntity.from_tag",
        }
    }
    with pytest.raises(ConfigError, match="could not be resolved"):
        cli.add_cli_arguments(parser, {"parse_args": new_args})



from __future__ import annotations

import itertools as it
import os
import sys
import warnings
from pathlib import Path
from typing import Protocol

import more_itertools as itx

from snakebids.io.console import in_interactive_session
from snakebids.paths import specs
from snakebids.paths._utils import BidsPathSpec, find_entity


class BidsFunction(Protocol):
    """Signature for functions returned by ``bids_factory``.

    See :func:`bids` for more details
    """

    def __call__(
        self,
        root: str | Path | None = None,
        *,
        datatype: str | None = None,
        prefix: str | None = None,
        suffix: str | None = None,
        extension: str | None = None,
        **entities: str | bool,
    ) -> str: ...


def _handle_subses_dir(
    root: str | Path | None = None,
    *,
    spec: BidsPathSpec,
    sub_dir_default: bool,
    ses_dir_default: bool,
    sub_dir: str | bool | None,
    ses_dir: str | bool | None,
    datatype: str | None = None,
    prefix: str | None = None,
    suffix: str | None = None,
    extension: str | None = None,
    **entities: str | bool,
) -> str | None:
    newspec = None
    warn_subses_dir = False

    if isinstance(ses_dir, bool):
        warn_subses_dir = True
        if ses_dir ^ ses_dir_default:
            if newspec is None:
                newspec = spec.copy()
            find_entity(newspec, "session")["dir"] = ses_dir

    if isinstance(sub_dir, bool):
        warn_subses_dir = True
        if sub_dir ^ sub_dir_default:
            if newspec is None:
                newspec = spec.copy()
            find_entity(newspec, "subject")["dir"] = sub_dir

    if warn_subses_dir:
        wrn_msg = (
            "include_session_dir and include_subject_dir are deprecated and "
            "will be removed in a future release. Builder functions without "
            "directories can be created using the bids_factory and spec "
            "functions:\n"
            "\tfrom snakebids.paths import bids_factory, specs\n"
            "\tbids_ses = bids_factory(specs.v0_0_0(subject_dir=False))\n"
            "\tbids_ses(...)\n"
        ).expandtabs(4)
        warnings.warn(wrn_msg, stacklevel=4)
    if newspec:
        return bids_factory(newspec)(
            root,
            datatype=datatype,
            prefix=prefix,
            suffix=suffix,
            extension=extension,
            **entities,
        )
    return None


def _get_entity_parser(aliases: dict[str, str]):
    def parse_entities(entities: dict[str, str | bool]) -> dict[str, str]:
        result: dict[str, str] = {}
        for entity, val in entities.items():
            # strip underscores from keys (needed so that users can use reserved
            # keywords by appending a _)
            stripped = entity.rstrip("_")
            unaliased = aliases.get(stripped, stripped)
            if unaliased in result:
                aliased = itx.nth(aliases, list(aliases.values()).index(unaliased))
                err = (
                    "Long and short names of an entity cannot be used in the same "
                    f"call to bids(): got '{aliased}' and '{unaliased}'"
                )
                raise ValueError(err)
            result[unaliased] = str(val)
        return result

    return parse_entities


def bids_factory(spec: BidsPathSpec, *, _implicit: bool = False) -> BidsFunction:
    """Generate bids functions according to the supplied spec.

    Parameters
    ----------
        spec
            Valid Bids Spec object
        _implicit
            Flag used internally to mark the default generated bids function. The
            resulting builder will warn when custom entities are used
    """
    order: list[str] = []
    dirs: set[str] = set()
    aliases: dict[str, str] = {}

    subject_dir_default = find_entity(spec, "subject").get("dir", False)
    session_dir_default = find_entity(spec, "session").get("dir", False)
    for entry in spec:
        tag = entry.get("tag", entry["entity"])
        order.append(tag)
        aliases[entry["entity"]] = tag
        if entry.get("dir"):
            dirs.add(tag)

    parse_entities = _get_entity_parser(aliases)

    def bids(
        root: str | Path | None = None,
        *,
        datatype: str | None = None,
        prefix: str | None = None,
        suffix: str | None = None,
        extension: str | None = None,
        **entities: str | bool,
    ) -> str:
        """Generate bids or bids-like paths.

        Path will have the general form::

            [root]/[sub-{{subject}}]/[ses-{{session}}]/
            [prefix]_[sub-{{subject}}]_[ses-{{session}}]_[{{key}}-{{val}}_...]_[suffix]

        If no arguments are specified, an empty string will be returned.

        Parameters
        ----------
        root
            Root folder to include in the path (e.g. ``results``)
        datatype
            Folder to include after sub-/ses- (e.g. ``anat``, ``dwi`` )
        prefix
            String to prepend to the file name. Useful for injecting custom entities at
            the front of the filename, e.g. ``tpl-{{tpl}}``
        suffix
            Suffix plus, optionally, the extension (e.g. ``T1w.nii.gz``)
        extension
            bids extension, beginning with ``.`` (e.g. ``.nii.gz``).  Typically
            shouldn't be specified manually: extensions should be listed along with the
            suffix.
        entities
            bids entities as keyword arguments paired with values (e.g. ``space="T1w"``
            for ``space-T1w``)
        """
        if (
            result := _handle_subses_dir(
                root,
                spec=spec,
                sub_dir_default=subject_dir_default,
                ses_dir_default=session_dir_default,
                sub_dir=entities.pop("include_subject_dir", None),
                ses_dir=entities.pop("include_session_dir", None),
                datatype=datatype,
                prefix=prefix,
                suffix=suffix,
                extension=extension,
                **entities,
            )
        ) is not None:
            return result

        if not any([entities, suffix, extension]) and any([datatype, prefix]):
            raise ValueError(
                "At least one of suffix, extension, or an entity must be "
                "supplied.\n\tGot only: "
                + " and ".join(
                    filter(
                        None,
                        (
                            f"datatype='{datatype}'" if datatype else None,
                            f"prefix='{prefix}'" if prefix else None,
                        ),
                    )
                )
            )

        parsed = parse_entities(entities)

        spec_parts: list[str] = []
        custom_parts: list[str] = []
        split: int = sys.maxsize + 1
        path_parts: list[str] = []

        if root:
            path_parts.append(str(root))
        if prefix:
            spec_parts.append(prefix)
        for entity in order:
            # Check for `*` first so that if user specifies an entity called `*` we
            # don't skip setting the split
            if entity == "*":
                split = len(spec_parts)
            elif value := parsed.pop(entity, None):
                spec_parts.append(f"{entity}-{value}")
                if entity in dirs:
                    path_parts.append(f"{entity}-{value}")
        for key, value in parsed.items():
            custom_parts.append(f"{key}-{value}")

        if datatype:
            path_parts.append(datatype)
        tail = [suffix] if suffix is not None else []
        path_parts.append(
            "_".join(
                it.chain(
                    spec_parts[:split],
                    custom_parts,
                    spec_parts[split:],
                    tail,
                )
            )
        )
        result = os.path.join(*path_parts)
        if extension is not None:
            result += extension

        if custom_parts and _implicit and not in_interactive_session():
            wrn_msg = (
                f"Path generated with unrecognized entities, and a snakebids spec has "
                "not been explicitly declared. This could break in future snakebids "
                "versions, as the default spec can be changed without warning.\n"
                f"\tpath = {result!r}\n"
                f"\tentities = {custom_parts!r}\n\n"
                "Please declare a spec using:\n"
                "\tfrom snakebids import set_bids_spec\n"
                f'\tset_bids_spec("{specs.LATEST}")\n'
            ).expandtabs(4)
            warnings.warn(wrn_msg, stacklevel=3)

        return result

    return bids



"""Script to generate a Snakebids project."""

import argparse
import re
import sys
from pathlib import Path

import copier
import more_itertools as itx
from colorama import Fore, Style

import snakebids
from snakebids.utils.utils import text_fold


def create_app(args: argparse.Namespace) -> None:
    """Implement the ``snakebids create`` command."""
    output = Path(args.output_dir).resolve()
    if not output.parent.exists():
        print(
            f"{Fore.RED}{Style.BRIGHT}{output.parent}{Style.RESET_ALL}{Fore.RED} does "
            f"not exist{Fore.RESET}",
            file=sys.stderr,
        )
        sys.exit(1)
    if not re.match(r"^[a-zA-Z_][a-zA-Z_0-9]*$", output.name):
        print(
            f"{Fore.RED}Output directory name {Style.BRIGHT}{output.name}"
            f"{Style.RESET_ALL}{Fore.RED} is not a valid python module name",
            file=sys.stderr,
        )
        sys.exit(1)

    data = {"app_full_name": output.name}

    if args.snakebids_version is not None:
        version = args.snakebids_version
        if ("@" not in version and ";" in version) or (
            "@" in version and " ;" in version
        ):
            print(
                f"{Fore.RED}Snakebids version may not specify markers{Style.RESET_ALL}",
                file=sys.stderr,
            )
            sys.exit(1)
        if (
            "@" in version
            and version[1:].lstrip().startswith("git+")
            and "@" in version[1:]
        ):
            print(
                f"{Fore.RED}Credentials and rev specifiers in git requirement "
                f"specifications are not supported{Style.RESET_ALL}",
                file=sys.stderr,
            )
            sys.exit(1)
        if version.strip().startswith("["):
            print(
                f"{Fore.RED}Snakebids version may not specify extras{Style.RESET_ALL}",
                file=sys.stderr,
            )
            sys.exit(1)

        data["snakebids_version"] = args.snakebids_version

    print(
        f"Creating Snakebids app at {Fore.GREEN}{output}{Fore.RESET}", file=sys.stderr
    )
    print(file=sys.stderr)
    try:
        copier.run_copy(
            str(Path(itx.first(snakebids.__path__), "project_template")),
            output,
            data=data,
            unsafe=True,
        )
    except KeyboardInterrupt:
        print(f"{Fore.RED}Aborted!{Fore.RESET}", file=sys.stderr)
        sys.exit(1)


def create_descriptor(args: argparse.Namespace) -> None:
    """Implement the ``snakebids boutiques`` command."""
    print("Boutiques descriptor generation is temporarily disabled", file=sys.stderr)
    sys.exit(1)


def gen_parser() -> argparse.ArgumentParser:
    """Generate the CLI parser for ``snakebids``."""
    parser = argparse.ArgumentParser(
        description="Perform administrative Snakebids tasks."
    )
    subparsers = parser.add_subparsers(required=True, dest="command")

    parser_create = subparsers.add_parser("create", help="Create a new Snakebids app.")
    parser_create.add_argument("output_dir", nargs="?", default=".")
    parser_create.add_argument(
        "--snakebids-version",
        default=None,
        metavar="VERSION_SPECIFIER",
        help=text_fold(
            """
            Specify snakebids version requirement. Supports either a valid version
            specifier (e.g. `>=x.x.x`, `==a.b.c`) or a url prepended with `@` (e.g. `@
            https://...`). Paths can be specified with `@ file:///absolute/path/...`.
            Markers and extras may not be specified.
            """
        ),
    )
    parser_create.set_defaults(func=create_app)

    parser_boutiques = subparsers.add_parser(
        "boutiques",
        help="Create a Boutiques descriptor for an existing Snakebids app.",
    )
    parser_boutiques.add_argument(
        "out_path",
        help="Path for the output Boutiques descriptor. Should be a .json file.",
        type=Path,
    )
    parser_boutiques.add_argument(
        "--app_dir",
        help="Location of the Snakebids app. Defaults to the current directory.",
        type=Path,
        default=".",
    )
    parser_boutiques.set_defaults(func=create_descriptor)

    return parser


def main() -> None:
    """Invoke Snakebids cli."""
    parser = gen_parser()
    args = parser.parse_args()
    args.func(args)



"""
This type stub file was generated by pyright.
"""

import sys
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    AnyStr,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    TextIO,
    Type,
    Union,
)

from pyfakefs.fake_filesystem import FakeFilesystem
from pyfakefs.helpers import AnyPath, AnyString

"""Fake implementations for different file objects.
"""
if TYPE_CHECKING: ...
AnyFileWrapper = Union[
    "FakeFileWrapper",
    "FakeDirWrapper",
    "StandardStreamWrapper",
    "FakePipeWrapper",
]
AnyFile = Union["FakeFile", "FakeDirectory"]

class FakeLargeFileIoException(Exception):
    """Exception thrown on unsupported operations for fake large files.
    Fake large files have a size with no real content.
    """

    def __init__(self, file_path: str) -> None: ...

class FakeFile:
    """Provides the appearance of a real file.

    Attributes currently faked out:
      * `st_mode`: user-specified, otherwise S_IFREG
      * `st_ctime`: the time.time() timestamp of the file change time (updated
        each time a file's attributes is modified).
      * `st_atime`: the time.time() timestamp when the file was last accessed.
      * `st_mtime`: the time.time() timestamp when the file was last modified.
      * `st_size`: the size of the file
      * `st_nlink`: the number of hard links to the file
      * `st_ino`: the inode number - a unique number identifying the file
      * `st_dev`: a unique number identifying the (fake) file system device
        the file belongs to
      * `st_uid`: always set to USER_ID, which can be changed globally using
            `set_uid`
      * `st_gid`: always set to GROUP_ID, which can be changed globally using
            `set_gid`

    .. note:: The resolution for `st_ctime`, `st_mtime` and `st_atime` in the
        real file system depends on the used file system (for example it is
        only 1s for HFS+ and older Linux file systems, but much higher for
        ext4 and NTFS). This is currently ignored by pyfakefs, which uses
        the resolution of `time.time()`.

        Under Windows, `st_atime` is not updated for performance reasons by
        default. pyfakefs never updates `st_atime` under Windows, assuming
        the default setting.
    """

    stat_types = ...
    def __init__(
        self,
        name: AnyStr,
        st_mode: int = ...,
        contents: Optional[AnyStr] = ...,
        filesystem: Optional[FakeFilesystem] = ...,
        encoding: Optional[str] = ...,
        errors: Optional[str] = ...,
        side_effect: Optional[Callable[[FakeFile], None]] = ...,
    ) -> None:
        """
        Args:
            name: Name of the file/directory, without parent path information
            st_mode: The stat.S_IF* constant representing the file type (i.e.
                stat.S_IFREG, stat.S_IFDIR), and the file permissions.
                If no file type is set (e.g. permission flags only), a
                regular file type is assumed.
            contents: The contents of the filesystem object; should be a string
                or byte object for regular files, and a dict of other
                FakeFile or FakeDirectory objects wih the file names as
                keys for FakeDirectory objects
            filesystem: The fake filesystem where the file is created.
            encoding: If contents is a unicode string, the encoding used
                for serialization.
            errors: The error mode used for encoding/decoding errors.
            side_effect: function handle that is executed when file is written,
                must accept the file object as an argument.
        """
        ...
    @property
    def byte_contents(self) -> Optional[bytes]:
        """Return the contents as raw byte array."""
        ...
    @property
    def contents(self) -> Optional[str]:
        """Return the contents as string with the original encoding."""
        ...
    @property
    def st_ctime(self) -> float:
        """Return the creation time of the fake file."""
        ...
    @st_ctime.setter
    def st_ctime(self, val: float) -> None:
        """Set the creation time of the fake file."""
        ...
    @property
    def st_atime(self) -> float:
        """Return the access time of the fake file."""
        ...
    @st_atime.setter
    def st_atime(self, val: float) -> None:
        """Set the access time of the fake file."""
        ...
    @property
    def st_mtime(self) -> float:
        """Return the modification time of the fake file."""
        ...
    @st_mtime.setter
    def st_mtime(self, val: float) -> None:
        """Set the modification time of the fake file."""
        ...
    def set_large_file_size(self, st_size: int) -> None:
        """Sets the self.st_size attribute and replaces self.content with None.

        Provided specifically to simulate very large files without regards
        to their content (which wouldn't fit in memory).
        Note that read/write operations with such a file raise
            :py:class:`FakeLargeFileIoException`.

        Args:
          st_size: (int) The desired file size

        Raises:
          OSError: if the st_size is not a non-negative integer,
                   or if st_size exceeds the available file system space
        """
        ...
    def is_large_file(self) -> bool:
        """Return `True` if this file was initialized with size
        but no contents.
        """
        ...
    def set_initial_contents(self, contents: AnyStr) -> bool:
        """Sets the file contents and size.
           Called internally after initial file creation.

        Args:
            contents: string, new content of file.

        Returns:
            True if the contents have been changed.

        Raises:
              OSError: if the st_size is not a non-negative integer,
                   or if st_size exceeds the available file system space
        """
        ...
    def set_contents(self, contents: AnyStr, encoding: Optional[str] = ...) -> bool:
        """Sets the file contents and size and increases the modification time.
        Also executes the side_effects if available.

        Args:
          contents: (str, bytes) new content of file.
          encoding: (str) the encoding to be used for writing the contents
                    if they are a unicode string.
                    If not given, the locale preferred encoding is used.

        Returns:
            True if the contents have been changed.

        Raises:
          OSError: if `st_size` is not a non-negative integer,
                   or if it exceeds the available file system space.
        """
        ...
    @property
    def size(self) -> int:
        """Return the size in bytes of the file contents."""
        ...
    @size.setter
    def size(self, st_size: int) -> None:
        """Resizes file content, padding with nulls if new size exceeds the
        old size.

        Args:
          st_size: The desired size for the file.

        Raises:
          OSError: if the st_size arg is not a non-negative integer
                   or if st_size exceeds the available file system space
        """
        ...
    @property
    def path(self) -> AnyStr:
        """Return the full path of the current object."""
        ...
    if sys.version_info >= (3, 12): ...
    def __getattr__(self, item: str) -> Any:
        """Forward some properties to stat_result."""
        ...
    def __setattr__(self, key: str, value: Any) -> None:
        """Forward some properties to stat_result."""
        ...
    def __str__(self) -> str: ...

class FakeNullFile(FakeFile):
    def __init__(self, filesystem: FakeFilesystem) -> None: ...
    @property
    def byte_contents(self) -> bytes: ...
    def set_initial_contents(self, contents: AnyStr) -> bool: ...

class FakeFileFromRealFile(FakeFile):
    """Represents a fake file copied from the real file system.

    The contents of the file are read on demand only.
    """

    def __init__(
        self,
        file_path: str,
        filesystem: FakeFilesystem,
        side_effect: Optional[Callable] = ...,
    ) -> None:
        """
        Args:
            file_path: Path to the existing file.
            filesystem: The fake filesystem where the file is created.

        Raises:
            OSError: if the file does not exist in the real file system.
            OSError: if the file already exists in the fake file system.
        """
        ...
    @property
    def byte_contents(self) -> Optional[bytes]: ...
    def set_contents(self, contents, encoding=...): ...
    def is_large_file(self):  # -> Literal[False]:
        """The contents are never faked."""
        ...

class FakeDirectory(FakeFile):
    """Provides the appearance of a real directory."""

    def __init__(
        self,
        name: str,
        perm_bits: int = ...,
        filesystem: Optional[FakeFilesystem] = ...,
    ) -> None:
        """
        Args:
            name:  name of the file/directory, without parent path information
            perm_bits: permission bits. defaults to 0o777.
            filesystem: if set, the fake filesystem where the directory
                is created
        """
        ...
    def set_contents(self, contents: AnyStr, encoding: Optional[str] = ...) -> bool: ...
    @property
    def entries(self) -> Dict[str, FakeFile]:
        """Return the list of contained directory entries."""
        ...
    @property
    def ordered_dirs(self) -> List[str]:
        """Return the list of contained directory entry names ordered by
        creation order.
        """
        ...
    def add_entry(self, path_object: FakeFile) -> None:
        """Adds a child FakeFile to this directory.

        Args:
            path_object: FakeFile instance to add as a child of this directory.

        Raises:
            OSError: if the directory has no write permission (Posix only)
            OSError: if the file or directory to be added already exists
        """
        ...
    def get_entry(self, pathname_name: str) -> AnyFile:
        """Retrieves the specified child file or directory entry.

        Args:
            pathname_name: The basename of the child object to retrieve.

        Returns:
            The fake file or directory object.

        Raises:
            KeyError: if no child exists by the specified name.
        """
        ...
    def remove_entry(self, pathname_name: str, recursive: bool = ...) -> None:
        """Removes the specified child file or directory.

        Args:
            pathname_name: Basename of the child object to remove.
            recursive: If True (default), the entries in contained directories
                are deleted first. Used to propagate removal errors
                (e.g. permission problems) from contained entries.

        Raises:
            KeyError: if no child exists by the specified name.
            OSError: if user lacks permission to delete the file,
                or (Windows only) the file is open.
        """
        ...
    @property
    def size(self) -> int:
        """Return the total size of all files contained
        in this directory tree.
        """
        ...
    @size.setter
    def size(self, st_size: int) -> None:
        """Setting the size is an error for a directory."""
        ...
    def has_parent_object(self, dir_object: FakeDirectory) -> bool:
        """Return `True` if dir_object is a direct or indirect parent
        directory, or if both are the same object."""
        ...
    def __str__(self) -> str: ...

class FakeDirectoryFromRealDirectory(FakeDirectory):
    """Represents a fake directory copied from the real file system.

    The contents of the directory are read on demand only.
    """

    def __init__(
        self,
        source_path: AnyPath,
        filesystem: FakeFilesystem,
        read_only: bool,
        target_path: Optional[AnyPath] = ...,
    ) -> None:
        """
        Args:
            source_path: Full directory path.
            filesystem: The fake filesystem where the directory is created.
            read_only: If set, all files under the directory are treated
                as read-only, e.g. a write access raises an exception;
                otherwise, writing to the files changes the fake files
                only as usually.
            target_path: If given, the target path of the directory,
                otherwise the target is the same as `source_path`.

        Raises:
            OSError: if the directory does not exist in the real file system
        """
        ...
    @property
    def entries(self) -> Dict[str, FakeFile]:
        """Return the list of contained directory entries, loading them
        if not already loaded."""
        ...
    @property
    def size(self) -> int: ...
    @size.setter
    def size(self, st_size: int) -> None: ...

class FakeFileWrapper:
    """Wrapper for a stream object for use by a FakeFile object.

    If the wrapper has any data written to it, it will propagate to
    the FakeFile object on close() or flush().
    """

    def __init__(
        self,
        file_object: FakeFile,
        file_path: AnyStr,
        update: bool,
        read: bool,
        append: bool,
        delete_on_close: bool,
        filesystem: FakeFilesystem,
        newline: Optional[str],
        binary: bool,
        closefd: bool,
        encoding: Optional[str],
        errors: Optional[str],
        buffering: int,
        raw_io: bool,
        is_stream: bool = ...,
    ) -> None: ...
    def __enter__(self) -> FakeFileWrapper:
        """To support usage of this fake file with the 'with' statement."""
        ...
    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """To support usage of this fake file with the 'with' statement."""
        ...
    def get_object(self) -> FakeFile:
        """Return the FakeFile object that is wrapped
        by the current instance.
        """
        ...
    def fileno(self) -> int:
        """Return the file descriptor of the file object."""
        ...
    def close(self) -> None:
        """Close the file."""
        ...
    @property
    def closed(self) -> bool:
        """Simulate the `closed` attribute on file."""
        ...
    def flush(self) -> None:
        """Flush file contents to 'disk'."""
        ...
    def update_flush_pos(self) -> None: ...
    def seek(self, offset: int, whence: int = ...) -> None:
        """Move read/write pointer in 'file'."""
        ...
    def tell(self) -> int:
        """Return the file's current position.

        Returns:
          int, file's current position in bytes.
        """
        ...
    def size(self) -> int:
        """Return the content size in bytes of the wrapped file."""
        ...
    def __getattr__(self, name: str) -> Any: ...
    def __iter__(self) -> Union[Iterator[str], Iterator[bytes]]: ...
    def __next__(self): ...

class StandardStreamWrapper:
    """Wrapper for a system standard stream to be used in open files list."""

    def __init__(self, stream_object: TextIO) -> None: ...
    def get_object(self) -> TextIO: ...
    def fileno(self) -> int:
        """Return the file descriptor of the wrapped standard stream."""
        ...
    def read(self, n: int = ...) -> bytes: ...
    def close(self) -> None:
        """We do not support closing standard streams."""
        ...
    def is_stream(self) -> bool: ...

class FakeDirWrapper:
    """Wrapper for a FakeDirectory object to be used in open files list."""

    def __init__(
        self,
        file_object: FakeDirectory,
        file_path: AnyString,
        filesystem: FakeFilesystem,
    ) -> None: ...
    def get_object(self) -> FakeDirectory:
        """Return the FakeFile object that is wrapped by the current
        instance."""
        ...
    def fileno(self) -> int:
        """Return the file descriptor of the file object."""
        ...
    def close(self) -> None:
        """Close the directory."""
        ...

class FakePipeWrapper:
    """Wrapper for a read or write descriptor of a real pipe object to be
    used in open files list.
    """

    def __init__(
        self, filesystem: FakeFilesystem, fd: int, can_write: bool, mode: str = ...
    ) -> None: ...
    def __enter__(self) -> FakePipeWrapper:
        """To support usage of this fake pipe with the 'with' statement."""
        ...
    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """To support usage of this fake pipe with the 'with' statement."""
        ...
    def get_object(self) -> None: ...
    def fileno(self) -> int:
        """Return the fake file descriptor of the pipe object."""
        ...
    def read(self, numBytes: int = ...) -> bytes:
        """Read from the real pipe."""
        ...
    def flush(self) -> None:
        """Flush the real pipe?"""
        ...
    def write(self, contents: bytes) -> int:
        """Write to the real pipe."""
        ...
    def close(self) -> None:
        """Close the pipe descriptor."""
        ...
    def readable(self) -> bool:
        """The pipe end can either be readable or writable."""
        ...
    def writable(self) -> bool:
        """The pipe end can either be readable or writable."""
        ...
    def seekable(self) -> bool:
        """A pipe is not seekable."""
        ...



from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING

from snakebids.utils.utils import entity_to_wildcard

if TYPE_CHECKING:
    from snakebids.paths._utils import BidsPathSpec, BidsPathSpecFile

TEMPLATE = '''
def {spec}(subject_dir: bool = True, session_dir: bool = True) -> BidsPathSpec:
    """{docstring}
    """
    ...
'''
DOCSTRING = """{description}

Parameters
----------
subject_dir
    If False, downstream path generator will not include the subject dir
    `sub-{{subject}}/*`
session_dir : bool, optional
    If False, downstream path generator will not include the session dir
    `*/ses-{{session}}/*`

"""

DEFAULT_DESCRIPTION = """Bids Spec {version}

Supply this to snakebids.bids_factory to construct a corresponding bids function
"""


def _wrap_template(template: str, length: int):
    def recurse(lines: list[str]) -> list[str]:
        line = lines[-1]
        if len(line) <= length:
            return lines
        i = line[length - 1 :: -1].index("_")
        return recurse([*lines[:-1], line[: length - i], 4 * " " + line[length - i :]])

    return "\n".join(recurse([template]))


def compile_example(spec: BidsPathSpec):
    # import within function to avoid circular import
    from snakebids.paths._factory import bids_factory  # noqa: PLC0415

    entities = [listing["entity"] for listing in spec]
    standard_entities = ("prefix", "datatype", "suffix", "extension")
    try:
        wild = entities.index("*")
    except ValueError:
        wild = -2
    template = bids_factory(spec)(
        **entity_to_wildcard(standard_entities),
        **entity_to_wildcard(e for e in entities if e != "*"),
    )
    search = (
        f"{spec[wild + 1].get('tag', entities[wild + 1])}-{{{entities[wild + 1]}}}_"
    )
    i = template.index(search)
    if wild < 0:
        i = len(search) + i
    return _wrap_template(template[:i] + "..._" + template[i:], 80)


def _import_docstring_parser():
    """Isolated import function that can be mocked in tests."""
    import docstring_parser as docstr  # noqa: PLC0415

    return docstr


def format_doc(spec: BidsPathSpecFile):
    if (description := spec.get("description")) is None:
        description = DEFAULT_DESCRIPTION.format(version=spec["version"])

    try:
        docstr = _import_docstring_parser()
    except ImportError:
        return description.strip()

    doc = docstr.parse(DOCSTRING.format(description=description.strip()))
    if doc.long_description:
        doc.long_description = (
            "\n\n".join(
                "\n".join(textwrap.wrap(para, 84))
                if not para.startswith("    ")
                else para
                for para in doc.long_description.split("\n\n")
            )
            + "\n\nFormatted as::\n\n    "
            + compile_example(spec["spec"])
        )

    return docstr.compose(doc)


def format_pyi(spec: BidsPathSpecFile):
    return TEMPLATE.format(
        spec=spec["version"].replace(".", "_"),
        docstring=format_doc(spec),
    )



# ruff: noqa: PLR2004
from __future__ import annotations

from math import inf

import pyparsing as pp
from hypothesis import assume, given
from hypothesis import strategies as st

import tests.strategies as sb_st
from snakebids.core.datasets import BidsComponent, BidsDataset
from snakebids.io.printing import format_zip_lists
from snakebids.types import ZipList


def zip_list_parser() -> pp.ParserElement:
    key = pp.quoted_string().set_results_name("key") + pp.Suppress(":")
    elided_list = (
        pp.Opt(pp.delimited_list(pp.quoted_string()))("left")
        + pp.Opt(",").suppress()
        + pp.Opt("...")("ellipse")
        + pp.Opt(pp.delimited_list(pp.quoted_string()))("right")
    )
    row = key + pp.Suppress("[") + elided_list + pp.Suppress("],")
    return pp.Suppress("{") + pp.Group(row)[1, ...] + pp.Suppress("}")


@given(zip_list=sb_st.zip_lists(max_entities=1, restrict_patterns=True))
def test_ellipses_appears_when_maxwidth_too_short(zip_list: ZipList):
    width = len(format_zip_lists(zip_list, tabstop=0).splitlines()[1])
    parsed = zip_list_parser().parse_string(
        format_zip_lists(zip_list, width - 1, tabstop=0)
    )
    assert "ellipse" in parsed[0]


@given(zip_list=sb_st.zip_lists(max_entities=1, restrict_patterns=True))
def test_no_ellipses_when_no_max_width(zip_list: ZipList):
    parsed = zip_list_parser().parse_string(format_zip_lists(zip_list, tabstop=0))
    assert "ellipse" not in parsed[0]


@given(zip_list=sb_st.zip_lists(max_entities=1, restrict_patterns=True))
def test_no_ellipses_when_max_width_long_enouth(zip_list: ZipList):
    width = len(format_zip_lists(zip_list, tabstop=0).splitlines()[1])
    parsed = zip_list_parser().parse_string(
        format_zip_lists(zip_list, width, tabstop=0)
    )
    assert "ellipse" not in parsed[0]


@given(
    zip_list=sb_st.zip_lists(
        max_entities=1, min_values=0, max_values=0, restrict_patterns=True
    )
)
def test_no_ellipses_appears_when_ziplist_empty(zip_list: ZipList):
    width = len(format_zip_lists(zip_list, tabstop=0).splitlines()[1])
    parsed = zip_list_parser().parse_string(
        format_zip_lists(zip_list, width - 1, tabstop=0)
    )
    assert "ellipse" not in parsed[0]


@given(
    zip_list=sb_st.zip_lists(
        min_values=1, max_values=4, max_entities=4, restrict_patterns=True
    ),
    width=st.integers(min_value=10, max_value=200),
)
def test_values_balanced_around_elision_correctly(zip_list: ZipList, width: int):
    parsed: pp.ParseResults = zip_list_parser().parse_string(
        format_zip_lists(zip_list, max_width=width, tabstop=0)
    )
    assert parsed
    assert parsed[0]
    nleft = len(parsed[0].get("left", []))  # type: ignore
    nright = len(parsed[0].get("right", []))  # type: ignore
    if "ellipse" in parsed[0]:
        assert nleft in {nright, nright + 1}
        assert nleft + nright == len(parsed[0]) - 2  # type: ignore
    else:
        assert nright == 0
        # nleft contains all elements in the first row except the key
        assert nleft == len(parsed[0]) - 1  # type: ignore

    for line in parsed[1:]:  # type: ignore
        assert len(line.get("left", [])) == nleft  # type: ignore
        assert len(line.get("right", [])) == nright  # type: ignore


class TestCorrectNumberOfLinesCreated:
    @given(
        zip_list=sb_st.zip_lists(
            min_values=0, max_values=1, max_entities=6, restrict_patterns=True
        ),
    )
    def test_in_zip_list(self, zip_list: ZipList):
        assert (
            len(format_zip_lists(zip_list, tabstop=0).splitlines()) == len(zip_list) + 2
        )

    @given(
        component=sb_st.bids_components(
            min_values=0, max_values=1, max_entities=6, restrict_patterns=True
        ),
    )
    def test_in_component(self, component: BidsComponent):
        assert (
            len(component.pformat(tabstop=0).splitlines())
            == len(component.entities) + 6
        )

    @given(dataset=sb_st.datasets())
    def test_in_dataset(self, dataset: BidsDataset):
        n_comps = len(dataset)
        n_entities = sum(len(comp.entities) for comp in dataset.values())
        assert (
            len(dataset.pformat(tabstop=0).splitlines()) == n_entities + 6 * n_comps + 2
        )


class TestIsValidPython:
    @given(
        zip_list=sb_st.zip_lists(restrict_patterns=True, min_values=0, min_entities=0)
    )
    def test_in_zip_list(self, zip_list: ZipList):
        assert eval(format_zip_lists(zip_list, inf)) == zip_list

    @given(component=sb_st.bids_components(restrict_patterns=True, min_values=0))
    def test_in_component(self, component: BidsComponent):
        assert eval(component.pformat(inf)) == component

    @given(dataset=sb_st.datasets())
    def test_in_dataset(self, dataset: BidsDataset):
        assert eval(dataset.pformat(inf)) == dataset


# this could also be tested for components and datasets, however, in those objects the
# path and name are allowed to be longer than the width, so finding the zip_list lines
# would prove more challenging than it's worth
@given(
    zip_list=sb_st.zip_lists(max_entities=1, restrict_patterns=True),
    width=st.integers(10, 100),
    tab=st.integers(0, 10),
)
def test_line_never_longer_than_max_width(zip_list: ZipList, width: int, tab: int):
    assume(width > tab + 10)
    formatted = format_zip_lists(zip_list, width, tab)
    parsed = zip_list_parser().parse_string(formatted)
    assume("left" in parsed[0])
    assert all(len(line) <= width for line in formatted.splitlines())


def get_indent_length(line: str):
    return len(line) - len(line.lstrip())


class TestIndentLengthMultipleOfTabStop:
    @given(
        zip_list=sb_st.zip_lists(restrict_patterns=True, min_values=0),
        tabstop=st.integers(1, 10),
    )
    def test_in_zip_list(self, zip_list: ZipList, tabstop: int):
        for line in format_zip_lists(zip_list, tabstop=tabstop).splitlines():
            assert get_indent_length(line) / tabstop in {0, 1}

    @given(
        component=sb_st.bids_components(
            min_values=0, max_values=1, restrict_patterns=True
        ),
        tabstop=st.integers(1, 10),
    )
    def test_in_component(self, component: BidsComponent, tabstop: int):
        for line in component.pformat(tabstop=tabstop).splitlines():
            assert get_indent_length(line) / tabstop in {0, 1, 2}

    @given(dataset=sb_st.datasets(), tabstop=st.integers(1, 10))
    def test_in_dataset(self, dataset: BidsDataset, tabstop: int):
        for line in dataset.pformat(tabstop=tabstop).splitlines():
            assert get_indent_length(line) / tabstop in {0, 1, 2, 3}


class TestMultipleLevelsOfIndentationUsed:
    @given(
        component=sb_st.bids_components(
            min_values=0, max_values=1, restrict_patterns=True
        ),
        tabstop=st.integers(1, 10),
    )
    def test_in_component(self, component: BidsComponent, tabstop: int):
        indents: set[int] = set()
        for line in component.pformat(tabstop=tabstop).splitlines():
            indents.add(get_indent_length(line))
        assert len(indents) == 3

    @given(dataset=sb_st.datasets(), tabstop=st.integers(1, 10))
    def test_in_dataset(self, dataset: BidsDataset, tabstop: int):
        indents: set[int] = set()
        for line in dataset.pformat(tabstop=tabstop).splitlines():
            indents.add(get_indent_length(line))
        assert len(indents) == 4



from __future__ import annotations

from pathlib import Path

from snakebids.paths._config import get_bids_func
from snakebids.paths.specs import LATEST


def bids(
    root: str | Path | None = None,
    *,
    datatype: str | None = None,
    prefix: str | None = None,
    suffix: str | None = None,
    extension: str | None = None,
    **entities: str | bool,
) -> str:
    """Generate bids or bids-like paths.

    File path is of the form::

        [root]/[sub-{subject}]/[ses-{session]/
        [prefix]_[sub-{subject}]_[ses-{session}]_[{key}-{val}_ ... ]_[suffix]

    If no arguments are specified, an empty string will be returned.

    Datatype and prefix may not be used in isolation, but must be given with
    another entity.

    BIDS paths are built based on specs, which are versioned for long-term stability.
    The latest version is ``<version>``. Information on its spec can be found at
    :func:`~snakebids.paths.specs.<version>`.

    .. warning::

        By default, ``bids()`` will always use the latest BIDS spec. This is unsafe for
        production environments, as the spec may be updated without warning, even on
        patch releases. These updates may change the path output by ``bids()``,
        resulting in breaking changes in downstream apps

        Production code should always explicitly set the spec version using
        :func:`~snakebids.set_bids_spec`::

            from snakebids import set_bids_spec
            set_bids_spec("<version>")

    Parameters
    ----------
    root
        Root folder to include in the path (e.g. ``results``)
    datatype
        Folder to include after sub-/ses- (e.g. ``anat``, ``dwi`` )
    prefix
        String to prepend to the file name. Useful for injecting custom entities at
        the front of the filename, e.g. ``tpl-{tpl}``
    suffix
        Suffix plus, optionally, the extension (e.g. ``T1w.nii.gz``)
    extension
        bids extension, beginning with ``.`` (e.g. ``.nii.gz``).  Typically
        shouldn't be specified manually: extensions should be listed along with the
        suffix.
    entities
        bids entities as keyword arguments paired with values (e.g. ``space="T1w"``
        for ``space-T1w``)


    Examples
    --------
    Below is a rule using bids naming for input and output::

        rule proc_img:
            input: 'sub-{subject}_T1w.nii.gz' output:
            'sub-{subject}_space-snsx32_desc-preproc_T1w.nii.gz'

    With bids() you can instead use::

            rule proc_img: input: bids(subject='{subject}',suffix='T1w.nii.gz')
            output: bids(
                subject='{subject}', space='snsx32', desc='preproc',
                suffix='T1w.nii.gz'
            )

    Note that here we are not actually using "functions as inputs" in snakemake,
    which would require a function definition with wildcards as the argument, and
    restrict to input/params, but bids() is being used simply to return a string.

    Also note that space, desc and suffix are NOT wildcards here, only {subject} is.
    This makes it easy to combine wildcards and non-wildcards with bids-like naming.

    However, you can still use bids() in a lambda function. This is especially
    useful if your wildcards are named the same as bids entities (e.g. {subject},
    {session}, {task} etc..)::

        rule proc_img:
            input: lambda wildcards: bids(**wildcards,suffix='T1w.nii.gz') output:
            bids(
                subject='{subject}', space='snsx32', desc='preproc',
                suffix='T1w.nii.gz'
            )

    Or another example where you may have many bids-like wildcards used in your
    workflow::

        rule denoise_func:
            input: lambda wildcards: bids(**wildcards, suffix='bold.nii.gz') output:
            bids(
                subject='{subject}', session='{session}', task='{task}',
                acq='{acq}', desc='denoise', suffix='bold.nii.gz'
            )

    In this example, all the wildcards will be determined from the output and passed
    on to bids() for inputs. The output filename will have a 'desc-denoise' flag
    added to it.

    Also note that even if you supply entities in a different order, the entities
    will be ordered based on the OrderedDict defined here. If entities not known are
    provided, they will be just be placed at the end (before the suffix), in the
    order you provide them in.
    """
    return get_bids_func()(
        root,
        datatype=datatype,
        prefix=prefix,
        suffix=suffix,
        extension=extension,
        **entities,
    )


assert bids.__doc__  # noqa: S101
bids.__doc__ = bids.__doc__.replace("<version>", LATEST)



"""
This type stub file was generated by pyright.
"""

from typing import Any, Optional

from ruamel.yaml.emitter import Emitter
from ruamel.yaml.representer import (
    BaseRepresenter,
    Representer,
    RoundTripRepresenter,
    SafeRepresenter,
)
from ruamel.yaml.resolver import BaseResolver, Resolver, VersionedResolver
from ruamel.yaml.serializer import Serializer

if False: ...
__all__ = ["BaseDumper", "SafeDumper", "Dumper", "RoundTripDumper"]

class BaseDumper(Emitter, Serializer, BaseRepresenter, BaseResolver):
    def __init__(
        self: Any,
        stream: StreamType,
        default_style: Any = ...,
        default_flow_style: Any = ...,
        canonical: Optional[bool] = ...,
        indent: Optional[int] = ...,
        width: Optional[int] = ...,
        allow_unicode: Optional[bool] = ...,
        line_break: Any = ...,
        encoding: Any = ...,
        explicit_start: Optional[bool] = ...,
        explicit_end: Optional[bool] = ...,
        version: Any = ...,
        tags: Any = ...,
        block_seq_indent: Any = ...,
        top_level_colon_align: Any = ...,
        prefix_colon: Any = ...,
    ) -> None: ...

class SafeDumper(Emitter, Serializer, SafeRepresenter, Resolver):
    def __init__(
        self,
        stream: StreamType,
        default_style: Any = ...,
        default_flow_style: Any = ...,
        canonical: Optional[bool] = ...,
        indent: Optional[int] = ...,
        width: Optional[int] = ...,
        allow_unicode: Optional[bool] = ...,
        line_break: Any = ...,
        encoding: Any = ...,
        explicit_start: Optional[bool] = ...,
        explicit_end: Optional[bool] = ...,
        version: Any = ...,
        tags: Any = ...,
        block_seq_indent: Any = ...,
        top_level_colon_align: Any = ...,
        prefix_colon: Any = ...,
    ) -> None: ...

class Dumper(Emitter, Serializer, Representer, Resolver):
    def __init__(
        self,
        stream: StreamType,
        default_style: Any = ...,
        default_flow_style: Any = ...,
        canonical: Optional[bool] = ...,
        indent: Optional[int] = ...,
        width: Optional[int] = ...,
        allow_unicode: Optional[bool] = ...,
        line_break: Any = ...,
        encoding: Any = ...,
        explicit_start: Optional[bool] = ...,
        explicit_end: Optional[bool] = ...,
        version: Any = ...,
        tags: Any = ...,
        block_seq_indent: Any = ...,
        top_level_colon_align: Any = ...,
        prefix_colon: Any = ...,
    ) -> None: ...

class RoundTripDumper(Emitter, Serializer, RoundTripRepresenter, VersionedResolver):
    def __init__(
        self,
        stream: StreamType,
        default_style: Any = ...,
        default_flow_style: Optional[bool] = ...,
        canonical: Optional[int] = ...,
        indent: Optional[int] = ...,
        width: Optional[int] = ...,
        allow_unicode: Optional[bool] = ...,
        line_break: Any = ...,
        encoding: Any = ...,
        explicit_start: Optional[bool] = ...,
        explicit_end: Optional[bool] = ...,
        version: Any = ...,
        tags: Any = ...,
        block_seq_indent: Any = ...,
        top_level_colon_align: Any = ...,
        prefix_colon: Any = ...,
    ) -> None: ...



from __future__ import annotations

import argparse
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any

import attrs
import more_itertools as itx

from snakebids import bidsapp
from snakebids.bidsapp.args import ArgumentGroups
from snakebids.exceptions import ConfigError
from snakebids.plugins.base import PluginBase


def _list_or_none(arg: str | Iterable[str] | None) -> list[str] | None:
    return arg if arg is None else list(itx.always_iterable(arg))


class _Derivative(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ):
        # the below two conditions are to cover the typing. They are not actually
        # encountered at this time in the code
        if values is None:  # pragma: no cover
            result = False
        elif isinstance(values, str):  # pragma: no cover
            result = [Path(values)]
        elif not len(values):
            result = True
        else:
            result = [Path(p) for p in values]
        setattr(namespace, self.dest, result)


@attrs.define
class BidsArgs(PluginBase):
    """Add basic BIDSApp arguments.

    Parameters
    ----------
    argument_group
        Specify title of the group to which arguments should be added
    bids_dir
        Indicate if bids_dir (first bids argument) should be defined
    output_dir
        Indicate if output_dir (second bids argument) should be defined
    analysis_level
        Indicate if output_dir (third bids argument) should be defined
    analysis_level_choices
        List of valid analysis levels
    analysis_level_choices_config
        Configuration key containing analysis level choices
    participant_label
        Indicate if ``participant_label`` should be defined. Used to filter specific
        subjects for processesing
    exclude_participant_label
        Indicate if ``exclude_participant_label`` should be defined. Used to excluded
        specific subjects from processesing
    derivatives
        Indicate if ``derivatives`` should be defined. Used to allow automatic
        derivative indexing or specify paths to derivatives.


    CLI Arguments
    ~~~~~~~~~~~~~
    All arguments are added by default, but can be disabled using their respective
    parameters. Additionally, arguments can be directly overridden by adding arguments
    to the parser before the plugin runs, using the following ``dests``:

    - ``bids_dir``: The input bids directory
    - ``output_dir``: The output bids directory
    - ``analysis_level``: The level of analysis to perform (usually ``participant`` or
      ``group``)
    - ``participant_label``: Collection of subject labels to include in analysis
    - ``exclude_participant_label``: Collection of subject labels to exclude from
      analysis
    - ``derivatives``: Collection of derivative folder paths to include in bids
      indexing.

    Note that only the ``dest`` field matters when overriding arguments, all other
    feature, including positional versus optional, are incidental. Other than adding
    arguments for the above ``dests``, ``BidsArgs`` makes no assumptions about these
    arguments. Other plugins, however, may expect to find data in the parser namespace
    or config corresponding to the ``dest`` names.

    .. warning::

        Overriding just one or two of the positional arguments may alter the order,
        preventing the app from being called correctly. Thus, if any of the positional
        args are being overridden, they all should be.


    Analysis levels
    ~~~~~~~~~~~~~~~
    Valid analysis levels can be set using the ``analysis_level_choices`` key or the
    ``analysis_level_choices_config`` key. If both are set, an error will be raised. If
    neither are set, the config will first be searched for the ``analysis_levels`` key.
    If this is not found, analysis levels will be set to ``["participant", "group"]``.
    """

    argument_group: str = ""
    bids_dir: bool = True
    output_dir: bool = True
    analysis_level: bool = True
    analysis_level_choices: list[str] | None = attrs.field(
        default=None, converter=_list_or_none
    )
    analysis_level_choices_config: str | None = None
    participant_label: bool = True
    exclude_participant_label: bool = True
    derivatives: bool = True

    def __eq__(self, other: Any):
        return isinstance(other, self.__class__)

    @bidsapp.hookimpl
    def add_cli_arguments(
        self,
        parser: argparse.ArgumentParser,
        argument_groups: ArgumentGroups,
        config: dict[str, Any],
    ):
        """Add arguments from config."""
        group = argument_groups[self.argument_group] if self.argument_group else parser

        if self.bids_dir:
            self.try_add_argument(
                group,
                dest="bids_dir",
                help="The directory with the input dataset formatted according "
                "to the BIDS standard.",
                type=Path,
            )
        if self.output_dir:
            self.try_add_argument(
                group,
                dest="output_dir",
                help="The directory where the output files "
                "should be stored. If you are running group level analysis "
                "this folder should be prepopulated with the results of the "
                "participant level analysis.",
                type=Path,
            )
        if self.analysis_level:
            if (
                self.analysis_level_choices is not None
                and self.analysis_level_choices_config is not None
            ):
                msg = (
                    "`analysis_level_choices` and `analysis_level_choices_config` "
                    "cannot be simultaneously defined"
                )
                raise ConfigError(msg)
            if self.analysis_level_choices is not None:
                analysis_levels = self.analysis_level_choices
            elif self.analysis_level_choices_config is not None:
                analysis_levels = config.get(self.analysis_level_choices_config)
            elif (analysis_levels := config.get("analysis_levels")) is None:
                analysis_levels = ["participant", "group"]
            self.try_add_argument(
                group,
                dest="analysis_level",
                help="Level of the analysis that will be performed.",
                choices=analysis_levels,
            )
        if self.participant_label:
            self.try_add_argument(
                group,
                "--participant-label",
                "--participant_label",
                help="The label(s) of the participant(s) that should be analyzed. The "
                "label corresponds to sub-<participant_label> from the BIDS sec "
                '(so it does not include "sub-"). If this parameter is not '
                "provided all subjects should be analyzed. Multiple "
                "participants can be specified with a space separated list.",
                dest="participant_label",
                metavar="LABEL",
                nargs="+",
            )
        if self.exclude_participant_label:
            self.try_add_argument(
                group,
                "--exclude-participant-label",
                "--exclude_participant_label",
                help="The label(s) of the participant(s) that should be excluded. The "
                "label corresponds to sub-<participant_label> from the BIDS spec "
                '(so it does not include "sub-"). If this parameter is not '
                "provided all subjects should be analyzed. Multiple "
                "participants can be specified with a space separated list.",
                metavar="LABEL",
                dest="exclude_participant_label",
                nargs="+",
            )
        if self.derivatives:
            self.try_add_argument(
                group,
                "--derivatives",
                help="Path(s) to a derivatives dataset, for folder(s) that contains "
                "multiple derivatives datasets",
                nargs="*",
                dest="derivatives",
                metavar="PATH",
                action=_Derivative,
                default=False,
            )



from __future__ import annotations

import importlib.metadata as impm
import re

import jinja2
import requests
from jinja2.ext import Extension


class SnakebidsVersionExtension(Extension):
    """Retrieve the latest snakebids version from pypi.

    Stores value in the ``snakebids_version`` global variable.
    """

    def __init__(self, env: jinja2.Environment):
        env.globals["curr_snakebids_version"] = self._lookup_version()  # type: ignore
        super().__init__(env)

    def _lookup_version(self):
        request = requests.get("https://pypi.org/pypi/snakebids/json", timeout=10)
        version_regex = re.compile(r"\d+\.\d+\.\d+")
        try:
            request.raise_for_status()
            version = request.json()["info"]["version"]
            version = version_regex.fullmatch(version).group()  # type: ignore
        except (requests.HTTPError, KeyError, AttributeError, TypeError):
            version = impm.version("snakebids")
            if version == "0.0.0" or not version_regex.fullmatch(version):
                return None
        return version



from __future__ import annotations

import errno
import json
from pathlib import Path
from typing import TYPE_CHECKING, Any

from snakebids.io.yaml import get_yaml_io

if TYPE_CHECKING:
    from _typeshed import StrPath


def write_config(
    config_file: StrPath, data: dict[str, Any], force_overwrite: bool = False
) -> None:
    """Write provided data as yaml or json to provided path.

    Output type is decided based on file suffix: .json -> JSON, .yaml,.yml -> YAML

    Parameters
    ----------
    config_file
        Path of yaml file
    data
        Data to format
    force_overwrite
        If True, force overwrite of already existing files, otherwise error out
    """
    config_file = Path(config_file)
    if (config_file.exists()) and not force_overwrite:
        err = FileExistsError(
            errno.EEXIST, f"'{config_file}' already exists", str(config_file)
        )
        raise err
    config_file.parent.mkdir(exist_ok=True)

    if config_file.suffix == ".json":
        with open(config_file, "w", encoding="utf-8") as f:
            json.dump(data, f, indent=4)
            return

    get_yaml_io().dump(data, config_file)



# flake8: noqa
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import datetime
import importlib.metadata as ilm

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../"))

# -- Project information -----------------------------------------------------

project = "Snakebids"
release = ilm.version("snakebids")
copyright = f"{datetime.date.today().year}, Ali R. Khan"
author = "Ali R. Khan"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinxarg.ext",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.autodoc",
    "sphinxcontrib.asciinema",
    "myst_parser",
    "sphinx_copybutton",
    "sphinx_reredirects",
    "sphinx_design",
]


myst_enable_extensions = [
    "attrs_block",
]
myst_enable_extensions = [
    "attrs_block",
    "attrs_inline",
    "tasklist",
    "deflist",
    "fieldlist",
]
myst_number_code_blocks = ["python", "yaml"]


napoleon_google_docstring = False
napoleon_numpy_docstring = True

autodoc_member_order = "bysource"
autodoc_typehints = "description"
autodoc_type_aliases = {
    "ZipList": "snakebids.types.ZipList",
    "ZipListLike": "snakebids.types.ZipListLike",
    "InputsConfig": "snakebids.types.InputsConfig",
}
autodoc_typehints_format = "short"
autosummary_generate = True
autoclass_content = "class"
autosummary_imported_members = True


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


master_doc = "index"

intersphinx_mapping = {
    "pybids": ("https://bids-standard.github.io/pybids/", None),
    "python": ("https://docs.python.org/3", None),
    "snakemake": ("https://snakemake.readthedocs.io/en/stable/", None),
    "pluggy": ("https://pluggy.readthedocs.io/en/stable", None),
}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_theme_options = {
    "source_repository": "https://github.com/khanlab/snakebids",
    "source_branch": "main",
    "source_directory": "docs/",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
# templates_path = ["_templates"]

html_css_files = ["css/typealiases.css"]

redirects = {
    "migration/0.5_to_0.6.md": "migration/0.5_to_0.8.html",
}

sphinxcontrib_asciinema_defaults = {
    "preload": 1,
    "rows": 24,
    "speed": 3,
}



# type: ignore
__submodules__ = ["hook", "run", "args"]

# <AUTOGEN_INIT>
import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

__all__ = ["ArgumentGroups", "ArgumentParserArgs", "SnakebidsConfig", "app", "hookimpl"]
# </AUTOGEN_INIT>



from __future__ import annotations

import argparse
import json
import logging
import subprocess as sp
import tempfile
from typing import Any

import attr

from snakebids import bidsapp
from snakebids.exceptions import SnakebidsPluginError
from snakebids.plugins.base import PluginBase
from snakebids.utils.utils import text_fold

_logger = logging.getLogger(__name__)


class InvalidBidsError(SnakebidsPluginError):
    """Error raised if an input BIDS dataset is invalid."""


@attr.define
class BidsValidator(PluginBase):
    """Perform validation of a BIDS dataset using the bids-validator.

    If the dataset is not valid according to the BIDS specifications, an
    InvalidBidsError is raised.

    An argument --skip-bids-validation is added to the CLI. This argument can be
    overridden by adding an argument with dest ``plugins.validator.skip`` before this
    plugin runs.

    Two entries are added to config:
    - ``"plugins.validator.skip"``: Indicates if validation was skipped
    - ``"plugins.validator.success"``: Indicates if validation succeeded

    Parameters
    ----------
    raise_invalid_bids : bool
        Flag to indicate whether InvalidBidsError should be raised if BIDS
        validation fails. Default to True.

    """

    raise_invalid_bids: bool = attr.field(default=True)

    PREFIX = "plugins.validator"

    def __eq__(self, other: Any):
        return isinstance(other, self.__class__)

    @bidsapp.hookimpl
    def add_cli_arguments(self, parser: argparse.ArgumentParser):
        """Add option to skip validation."""
        self.try_add_argument(
            parser,
            "--skip-bids-validation",
            dest="skip",
            action="store_true",
            default=False,
            help=text_fold(
                """
                Skip validation of BIDS dataset. BIDS validation is performed by default
                using the bids-validator plugin (if installed/enabled) or with the
                pybids validator implementation (if bids-validator is not
                installed/enabled).
                """
            ),
        )

    @bidsapp.hookimpl
    def finalize_config(self, config: dict[str, Any]) -> None:
        """Perform BIDS validation of dataset.

        Raises
        ------
        InvalidBidsError
            Raised when the input BIDS directory does not pass validation with
            the bids-validator
        """
        # Skip bids validation
        if config["plugins.validator.skip"]:
            return

        validator_config_dict = {"ignoredFiles": ["/participants.tsv"]}

        with tempfile.NamedTemporaryFile(mode="w+", suffix=".json") as temp:
            temp.write(json.dumps(validator_config_dict))
            temp.flush()
            try:
                sp.check_call(["bids-validator", config["bids_dir"], "-c", temp.name])

                # If successfully bids-validation performed
                config["plugins.validator.success"] = True
            except FileNotFoundError:
                # If the bids-validator call can't be made
                config["plugins.validator.success"] = False
                _logger.warning(
                    "Missing bids-validator installation - falling back to pybids "
                    "validation."
                )
            # Any other bids-validator error
            except sp.CalledProcessError as err:
                config["plugins.validator.success"] = False
                if self.raise_invalid_bids:
                    raise InvalidBidsError from err



# ruff: noqa: PLR2004
from __future__ import annotations

import filecmp
import functools as ft
import itertools as it
import keyword
import os
import re
import shutil
import sys
import tempfile
import warnings
from collections import defaultdict
from collections.abc import Iterable
from pathlib import Path, PosixPath
from typing import Any, Literal, NamedTuple, TypedDict, TypeVar

import attrs
import more_itertools as itx
import pytest
from bids import BIDSLayout
from hypothesis import HealthCheck, assume, example, given, settings
from hypothesis import strategies as st
from pytest_mock import MockerFixture

from snakebids.core import input_generation
from snakebids.core._querying import PostFilter, UnifiedFilter, get_matching_files
from snakebids.core.datasets import BidsComponent, BidsDataset
from snakebids.core.input_generation import (
    _all_custom_paths,
    _gen_bids_layout,
    _get_components,
    _is_local_relative,
    _normalize_database_args,
    _parse_bids_path,
    _parse_custom_path,
    generate_inputs,
)
from snakebids.exceptions import ConfigError, PybidsError, RunError
from snakebids.paths._presets import bids
from snakebids.snakemake_compat import expand as sb_expand
from snakebids.types import InputsConfig
from snakebids.utils.containers import MultiSelectDict
from snakebids.utils.utils import DEPRECATION_FLAG, BidsEntity, BidsParseError
from tests import strategies as sb_st
from tests.helpers import (
    Benchmark,
    BidsListCompare,
    allow_function_scoped,
    create_dataset,
    create_snakebids_config,
    get_bids_path,
    get_zip_list,
    mock_data,
    reindex_dataset,
)

T = TypeVar("T")


def _not_deprecated(s: str):
    return not (s.startswith(DEPRECATION_FLAG) and s.endswith(DEPRECATION_FLAG))


class TestNormalizeDatabaseArgs:
    @given(
        pybidsdb_dir=st.text().filter(_not_deprecated) | st.none(),
        pybidsdb_reset=st.booleans(),
    )
    def test_normal_calls_give_no_warnings(
        self, pybidsdb_dir: str | None, pybidsdb_reset: bool
    ):
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            _normalize_database_args(pybidsdb_dir, pybidsdb_reset, None, None)

    @given(
        pybidsdb_dir=st.text().filter(_not_deprecated) | st.none(),
        pybidsdb_reset=st.booleans(),
        pybids_database_dir=st.text().filter(_not_deprecated),
    )
    def test_old_dir_param_gives_warning(
        self, pybidsdb_dir: str | None, pybidsdb_reset: bool, pybids_database_dir: str
    ):
        with pytest.warns(UserWarning, match="`pybids_database_dir`"):
            _normalize_database_args(
                pybidsdb_dir, pybidsdb_reset, pybids_database_dir, None
            )

    @given(
        pybidsdb_dir=st.text().filter(_not_deprecated) | st.none(),
        pybidsdb_reset=st.booleans(),
        pybids_database_reset=st.booleans(),
    )
    def test_old_reset_param_gives_warning(
        self,
        pybidsdb_dir: str | None,
        pybidsdb_reset: bool,
        pybids_database_reset: bool,
    ):
        with pytest.warns(UserWarning, match="`pybids_reset_database`"):
            _normalize_database_args(
                pybidsdb_dir, pybidsdb_reset, None, pybids_database_reset
            )

    @given(
        pybids_database_dir=st.text().filter(_not_deprecated),
        pybids_database_reset=st.booleans(),
    )
    def test_old_params_passed_on_to_new(
        self,
        pybids_database_dir: str,
        pybids_database_reset: bool,
    ):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            assert (
                pybids_database_dir,
                pybids_database_reset,
            ) == _normalize_database_args(
                None, None, pybids_database_dir, pybids_database_reset
            )

    @given(
        pybidsdb_reset=st.booleans() | st.none(),
        pybids_database_reset=st.booleans() | st.none(),
    )
    def test_second_return_never_none(
        self,
        pybidsdb_reset: bool | None,
        pybids_database_reset: bool | None,
    ):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=r"The parameter `pybids_reset_database` in generate_inputs\(\) "
                "is deprecated and will be removed in the next release. To reset the "
                "pybids database, use the `pybidsdb_reset` parameter instead",
                category=UserWarning,
            )
            assert (
                _normalize_database_args(
                    None,
                    pybidsdb_reset,
                    None,
                    pybids_database_reset,
                )[1]
                is not None
            )

    @given(pybidsdb_dir=st.text().filter(_not_deprecated))
    def test_deprecated_dir_raises_warning(self, pybidsdb_dir: str):
        pybidsdb_dir_ = DEPRECATION_FLAG + pybidsdb_dir + DEPRECATION_FLAG
        with pytest.warns(UserWarning, match="`pybids_db_dir`"):
            assert (
                _normalize_database_args(pybidsdb_dir_, None, None, None)[0]
                == pybidsdb_dir
            )

    @given(pybidsdb_reset=st.booleans())
    def test_deprecated_reset_raises_warning(self, pybidsdb_reset: bool):
        pybidsdb_reset_ = DEPRECATION_FLAG + str(int(pybidsdb_reset)) + DEPRECATION_FLAG
        with pytest.warns(UserWarning, match="`pybids_db_reset`"):
            assert (
                _normalize_database_args(None, pybidsdb_reset_, None, None)[1]
                == pybidsdb_reset
            )

    @given(pybidsdb_reset=st.text().filter(_not_deprecated))
    def test_non_deprecated_text_in_reset_raises_error(self, pybidsdb_reset: bool):
        with pytest.raises(TypeError):
            _normalize_database_args(None, pybidsdb_reset, None, None)


def test_regex_search_removed_from_filters():
    assert not len(UnifiedFilter.from_filter_dict({"regex_search": "foo"}).prefilters)


@given(
    filters=st.dictionaries(st.text(), st.text() | st.booleans() | st.lists(st.text()))
)
def test_get_matching_files_skips_get_when_empty_prefilter(filters: dict[str, Any]):
    assert (
        get_matching_files(
            ...,  # type: ignore
            UnifiedFilter.from_filter_dict({**filters, "foo": []}),
        )
        == []
    )


def test_attribute_errors_from_pybids_qualified_and_raised():
    with pytest.raises(PybidsError, match="Pybids has encountered a problem"):
        get_matching_files(..., UnifiedFilter.from_filter_dict({}))  # type: ignore


class TestFilterBools:
    def disambiguate_components(self, dataset: BidsDataset):
        assert len(dataset) == 2
        comp1, comp2 = dataset.values()
        return sorted([comp1, comp2], key=lambda comp: len(comp.entities.keys()))

    def get_extra_entity(self, dataset: BidsDataset) -> str:
        def set_difference(set_: set[str], comp: BidsComponent) -> set[str]:
            return set_.symmetric_difference(comp.entities.keys())

        return itx.one(
            ft.reduce(
                set_difference,
                dataset.values(),
                set[str](),
            )
        )

    @settings(
        deadline=800,
        suppress_health_check=[
            HealthCheck.function_scoped_fixture,
            HealthCheck.too_slow,
        ],
    )
    @given(dataset=sb_st.datasets())
    def test_ambiguous_paths_with_extra_entities_leads_to_error(
        self, tmpdir: Path, dataset: BidsDataset
    ):
        root = tempfile.mkdtemp(dir=tmpdir)
        create_dataset(root, dataset)
        shorter, _ = self.disambiguate_components(dataset)
        pybids_inputs: InputsConfig = {
            shorter.input_name: {
                "wildcards": [
                    BidsEntity.from_tag(wildcard).entity
                    for wildcard in shorter.input_wildcards
                ],
            }
        }

        with pytest.raises(ConfigError):
            generate_inputs(root, pybids_inputs)

    @settings(
        deadline=800,
        suppress_health_check=[
            HealthCheck.function_scoped_fixture,
            HealthCheck.too_slow,
        ],
    )
    @given(dataset=sb_st.datasets())
    def test_ambiguous_paths_with_missing_entity_leads_to_error(
        self, tmpdir: Path, dataset: BidsDataset
    ):
        root = tempfile.mkdtemp(dir=tmpdir)
        create_dataset(root, dataset)
        _, longer = self.disambiguate_components(dataset)
        pybids_inputs: InputsConfig = {
            longer.input_name: {
                "wildcards": [
                    BidsEntity.from_tag(wildcard).entity
                    for wildcard in longer.input_wildcards
                ],
            }
        }

        with pytest.raises(ConfigError):
            generate_inputs(root, pybids_inputs)

    @settings(
        deadline=800,
        suppress_health_check=[
            HealthCheck.function_scoped_fixture,
            HealthCheck.too_slow,
        ],
    )
    @given(dataset=sb_st.datasets(unique=True))
    def test_entity_excluded_when_filter_false(
        self, tmpdir: Path, dataset: BidsDataset
    ):
        root = tempfile.mkdtemp(dir=tmpdir)
        create_dataset(root, dataset)
        shorter, _ = self.disambiguate_components(dataset)
        extra_entity = self.get_extra_entity(dataset)
        pybids_inputs: InputsConfig = {
            shorter.input_name: {
                "wildcards": [
                    BidsEntity.from_tag(wildcard).entity
                    for wildcard in shorter.input_wildcards
                ],
                "filters": {BidsEntity.from_tag(extra_entity).entity: False},
            }
        }

        expected = BidsDataset(
            {
                shorter.input_name: attrs.evolve(
                    shorter, path=os.path.join(root, shorter.input_path)
                )
            }
        )

        data = generate_inputs(root, pybids_inputs)
        assert data == expected

    @example(
        dataset=BidsDataset(
            {
                "0": BidsComponent(
                    name="0",
                    path="ce-{ce}_space-{space}",
                    zip_lists={"ce": ["0"], "space": ["0"]},
                ),
                "1": BidsComponent(name="1", path="ce-{ce}", zip_lists={"ce": ["0"]}),
            }
        )
    )
    @example(
        dataset=BidsDataset(
            {
                "1": BidsComponent(
                    name="1",
                    path="sub-{subject}/{datatype}/sub-{subject}",
                    zip_lists={"subject": ["0"], "datatype": ["anat"]},
                ),
                "0": BidsComponent(
                    name="0",
                    path="sub-{subject}/sub-{subject}",
                    zip_lists={"subject": ["0"]},
                ),
            }
        )
    )
    @example(
        dataset=BidsDataset(
            {
                "1": BidsComponent(
                    name="1",
                    path="sub-{subject}/sub-{subject}_{suffix}.foo",
                    zip_lists={"subject": ["0"], "suffix": ["bar"]},
                ),
                "0": BidsComponent(
                    name="0",
                    path="{suffix}.foo",
                    zip_lists={"suffix": ["bar"]},
                ),
            }
        )
    )
    @settings(
        deadline=800,
        suppress_health_check=[
            HealthCheck.function_scoped_fixture,
            HealthCheck.too_slow,
        ],
    )
    @given(dataset=sb_st.datasets(unique=True))
    def test_entity_excluded_when_filter_true(self, tmpdir: Path, dataset: BidsDataset):
        root = tempfile.mkdtemp(dir=tmpdir)
        create_dataset(root, dataset)
        _, longer = self.disambiguate_components(dataset)
        extra_entity = self.get_extra_entity(dataset)
        pybids_inputs: InputsConfig = {
            longer.input_name: {
                "wildcards": [
                    BidsEntity.from_tag(wildcard).entity
                    for wildcard in longer.input_wildcards
                ],
                "filters": {BidsEntity.from_tag(extra_entity).entity: True},
            }
        }

        expected = BidsDataset(
            {
                longer.input_name: attrs.evolve(
                    longer, path=os.path.join(root, longer.input_path)
                )
            }
        )

        data = generate_inputs(root, pybids_inputs)
        assert data == expected

    @allow_function_scoped
    @given(
        template=sb_st.bids_components(
            name="template", restrict_patterns=True, unique=True, extra_entities=False
        ),
        entity=sb_st.bids_entity(path_safe=True),
        data=st.data(),
    )
    def test_text_filter_selects_paths_when_in_list_with_false(
        self,
        tmpdir: Path,
        template: BidsComponent,
        entity: BidsEntity,
        data: st.DataObject,
    ):
        """Also test here that the filters can pick out a desired value from a decoy"""

        def add_entity(component: BidsComponent, entity: str, value: str):
            return get_bids_path(component.wildcards, **{entity: value})

        assume(entity.wildcard not in template.wildcards)
        root = tempfile.mkdtemp(dir=tmpdir)

        target = data.draw(sb_st.bids_value(entity.match))
        decoy = data.draw(sb_st.bids_value(entity.match))
        assume(target != decoy)
        # This is a brute-force way of dealing with the fact that values for `run` are
        # converted into int by pybids before querying. So just prevent the decoy from
        # ever looking like the same number as target
        if target.isdecimal() and decoy.isdecimal():
            assume(int(target) != int(decoy))
        dataset = BidsDataset.from_iterable(
            [
                attrs.evolve(
                    template,
                    name="target",
                    path=os.path.join(
                        root, add_entity(template, entity.wildcard, target)
                    ),
                ),
                attrs.evolve(
                    template,
                    name="decoy",
                    path=os.path.join(
                        root, add_entity(template, entity.wildcard, decoy)
                    ),
                ),
            ]
        )
        create_dataset("", dataset)
        pybids_inputs: InputsConfig = {
            "target": {
                "wildcards": [
                    BidsEntity.from_tag(wildcard).entity
                    for wildcard in template.wildcards
                ],
                "filters": {entity.entity: [target, False]},
            }
        }

        result = generate_inputs(root, pybids_inputs)
        assert result == BidsDataset({"target": dataset["target"]})

    @allow_function_scoped
    @given(
        template=sb_st.bids_components(
            name="template", restrict_patterns=True, unique=True, extra_entities=False
        ),
        entity=sb_st.bids_entity(path_safe=True),
        data=st.data(),
    )
    def test_filter_false_in_list_selects_paths(
        self,
        tmpdir: Path,
        template: BidsComponent,
        entity: BidsEntity,
        data: st.DataObject,
    ):
        def add_entity(component: BidsComponent, entity: str, value: str):
            return get_bids_path(component.wildcards, **{entity: value})

        assume(entity.wildcard not in template.wildcards)
        target = data.draw(sb_st.bids_value(entity.match))
        decoy = data.draw(sb_st.bids_value(entity.match))
        assume(target != decoy)
        # See note in test_filter_works_when_false_in_list
        if target.isdecimal() and decoy.isdecimal():
            assume(int(target) != int(decoy))

        root = tempfile.mkdtemp(dir=tmpdir)
        dataset = BidsDataset.from_iterable(
            [
                attrs.evolve(
                    template,
                    path=os.path.join(root, template.path),
                ),
                attrs.evolve(
                    template,
                    name="decoy",
                    path=os.path.join(
                        root, add_entity(template, entity.wildcard, decoy)
                    ),
                ),
            ]
        )
        create_dataset("", dataset)
        pybids_inputs: InputsConfig = {
            "template": {
                "wildcards": [
                    BidsEntity.from_tag(wildcard).entity
                    for wildcard in template.wildcards
                ],
                "filters": {entity.entity: [target, False]},
            }
        }

        result = generate_inputs(root, pybids_inputs)
        assert result == BidsDataset({"template": dataset["template"]})


class TestFilterMethods:
    @example(
        component=BidsComponent(
            name="template",
            path="sub-{subject}/ses-{session}/sub-{subject}_ses-{session}",
            zip_lists={
                "session": ["0A", "0a"],
                "subject": ["0", "00"],
            },
        ),
        data=mock_data(["0a"]),
    )
    @example(
        component=BidsComponent(
            name="template",
            path="sub-{subject}/sub-{subject}_mt-{mt}",
            zip_lists={
                "subject": ["0", "00"],
                "mt": ["on", "on"],
            },
        ),
        data=mock_data(["0"]),
    )
    @given(
        component=sb_st.bids_components(
            name="template",
            min_entities=2,
            restrict_patterns=True,
            unique=True,
            extra_entities=False,
            # pybids bug prevents regex matching from working properly with extension
            blacklist_entities=["extension"],
        ),
        data=st.data(),
    )
    @settings(
        deadline=None,
        suppress_health_check=[
            HealthCheck.function_scoped_fixture,
            HealthCheck.too_slow,
        ],
    )
    def test_regex_match_selects_paths(
        self, tmpdir: Path, component: BidsComponent, data: st.DataObject
    ):
        root = tempfile.mkdtemp(dir=tmpdir)
        entity = itx.first(component.entities)
        selection = data.draw(
            st.lists(
                st.sampled_from(component.entities[entity]), unique=True, min_size=1
            )
        )
        dataset = BidsDataset.from_iterable(
            [attrs.evolve(component, path=os.path.join(root, component.path))]
        )
        create_dataset("", dataset)
        pybids_inputs: InputsConfig = {
            "template": {
                "wildcards": [
                    BidsEntity.from_tag(wildcard).entity
                    for wildcard in component.wildcards
                ],
                "filters": {
                    BidsEntity.from_tag(entity).entity: {
                        "match": "|".join(re.escape(sel) for sel in selection)
                    }
                },
            }
        }
        result = generate_inputs(root, pybids_inputs)
        assert result == BidsDataset(
            {"template": dataset["template"].filter(**{entity: selection})}
        )

    @given(
        component=sb_st.bids_components(
            name="template",
            min_entities=2,
            restrict_patterns=True,
            unique=True,
            extra_entities=False,
            # only specified entities work with free text
            whitelist_entities=[
                "subject",
                "session",
                "sample",
                "task",
                "acquisition",
                "ceagent",
                "staning",
                "tracer",
                "reconstruction",
                "direction",
                "proc",
                "modality",
                "recording",
                "space",
                "split",
                "atlas",
                "roi",
                "label",
                "from",
                "to",
                "res",
                "den",
                "model",
                "subset",
                "desc",
                "tracksys",
            ],
        ),
    )
    @settings(
        deadline=None,
        suppress_health_check=[
            HealthCheck.function_scoped_fixture,
            HealthCheck.too_slow,
        ],
    )
    def test_regex_search_selects_paths(self, tmpdir: Path, component: BidsComponent):
        root = tempfile.mkdtemp(dir=tmpdir)
        entity = itx.first(component.entities)
        assume(f"prefix{component[entity][0]}suffix" not in component.entities[entity])
        zip_lists = {
            ent: (
                [*value, f"prefix{value[0]}suffix"]
                if ent is entity
                else [*value, value[0]]
            )
            for ent, value in component.zip_lists.items()
        }
        dataset = BidsDataset.from_iterable(
            [
                attrs.evolve(
                    component,
                    path=os.path.join(root, component.path),
                    zip_lists=zip_lists,
                )
            ]
        )
        create_dataset("", dataset)
        pybids_inputs: InputsConfig = {
            "template": {
                "wildcards": [
                    BidsEntity.from_tag(wildcard).entity
                    for wildcard in component.wildcards
                ],
                "filters": {
                    BidsEntity.from_tag(entity).entity: {
                        "search": "|".join(
                            re.escape(val) for val in component.entities[entity]
                        )
                    }
                },
            }
        }
        result = generate_inputs(root, pybids_inputs)
        assert result == BidsDataset({"template": dataset["template"]})

    @example(
        component=BidsComponent(
            name="template",
            path="sub-{subject}/sub-{subject}_mt-{mt}",
            zip_lists={
                "subject": ["0", "00"],
                "mt": ["on", "on"],
            },
        ),
        data=mock_data(["0"]),
    )
    @given(
        component=sb_st.bids_components(
            name="template",
            min_entities=2,
            restrict_patterns=True,
            unique=True,
            extra_entities=False,
        ),
        data=st.data(),
    )
    @settings(
        deadline=None,
        suppress_health_check=[
            HealthCheck.function_scoped_fixture,
            HealthCheck.too_slow,
        ],
    )
    def test_get_method_selects_via_direct_matching(
        self, tmpdir: Path, component: BidsComponent, data: st.DataObject
    ):
        root = tempfile.mkdtemp(dir=tmpdir)
        entity = itx.first(component.entities)
        selection = data.draw(
            st.lists(
                st.sampled_from(component.entities[entity]), unique=True, min_size=1
            )
        )
        dataset = BidsDataset.from_iterable(
            [attrs.evolve(component, path=os.path.join(root, component.path))]
        )
        create_dataset("", dataset)
        pybids_inputs: InputsConfig = {
            "template": {
                "wildcards": [
                    BidsEntity.from_tag(wildcard).entity
                    for wildcard in component.wildcards
                ],
                "filters": {BidsEntity.from_tag(entity).entity: {"get": selection}},
            }
        }
        result = generate_inputs(root, pybids_inputs)
        assert result == BidsDataset(
            {"template": dataset["template"].filter(**{entity: selection})}
        )

    @given(
        component=sb_st.bids_components(
            name="template",
            min_entities=2,
            max_entities=2,
            # Again, extension doesn't work with regex. run causes problem with
            # numeral strings
            blacklist_entities={"extension", "run"},
            min_values=2,
            restrict_patterns=True,
            unique=True,
            cull=False,
            extra_entities=False,
        ),
    )
    @settings(
        deadline=None,
        suppress_health_check=[
            HealthCheck.function_scoped_fixture,
            HealthCheck.too_slow,
        ],
    )
    def test_combining_match_and_get_selects_correct_paths(
        self, tmpdir: Path, component: BidsComponent
    ):
        root = tempfile.mkdtemp(dir=tmpdir)
        entity1 = itx.first(component.entities)
        entity2 = itx.nth_or_last(component.entities, 1)
        dataset = BidsDataset.from_iterable(
            [attrs.evolve(component, path=os.path.join(root, component.path))]
        )
        create_dataset("", dataset)
        pybids_inputs: InputsConfig = {
            "template": {
                "wildcards": [
                    BidsEntity.from_tag(wildcard).entity
                    for wildcard in component.wildcards
                ],
                "filters": {
                    BidsEntity.from_tag(entity1).entity: {"get": component[entity1][0]},
                    BidsEntity.from_tag(entity2).entity: {
                        "match": re.escape(component[entity2][0])
                    },
                },
            }
        }
        result = generate_inputs(root, pybids_inputs)
        assert len(itx.first(itx.first(result.values()).zip_lists.values())) == 1

    @given(
        methods=st.lists(
            st.sampled_from(["match", "get", "search"]), unique=True, min_size=2
        )
    )
    @allow_function_scoped
    def test_filter_with_multiple_methods_raises_error(
        self, tmpdir: Path, methods: list[str]
    ):
        dataset = BidsDataset.from_iterable(
            [BidsComponent(zip_lists={}, name="template", path=str(tmpdir))]
        )
        create_dataset("", dataset)
        pybids_inputs: InputsConfig = {
            "template": {
                "filters": {
                    "foo": dict.fromkeys(methods, "foo")  # type: ignore
                },
            }
        }
        with pytest.raises(ConfigError, match="may not have more than one key"):
            generate_inputs(tmpdir, pybids_inputs)

    @pytest.mark.disable_fakefs(True)
    def test_filter_with_no_methods_raises_error(self, tmpdir: Path):
        dataset = BidsDataset.from_iterable(
            [BidsComponent(zip_lists={}, name="template", path=str(tmpdir))]
        )
        create_dataset("", dataset)
        pybids_inputs: InputsConfig = {
            "template": {
                "filters": {"foo": {}},
            }
        }
        with pytest.raises(ConfigError, match="was not given any keys"):
            generate_inputs(tmpdir, pybids_inputs)

    @given(
        method=st.text().filter(lambda s: s not in {"get", "match", "search"}),
    )
    @settings(
        deadline=None,
        suppress_health_check=[
            HealthCheck.function_scoped_fixture,
            HealthCheck.too_slow,
        ],
    )
    def test_filter_with_invalid_method_raises_error(self, tmpdir: Path, method: str):
        dataset = BidsDataset.from_iterable(
            [BidsComponent(zip_lists={}, name="template", path=str(tmpdir))]
        )
        create_dataset("", dataset)
        pybids_inputs: InputsConfig = {
            "template": {
                "filters": {"foo": {method: []}},  # type: ignore
            }
        }
        with pytest.raises(ConfigError, match="Invalid query method specified"):
            generate_inputs(tmpdir, pybids_inputs)


@settings(
    deadline=800,
    suppress_health_check=[
        HealthCheck.function_scoped_fixture,
        HealthCheck.too_slow,
    ],
    max_examples=1,
)
@given(dataset=sb_st.datasets_one_comp(unique=True))
def test_duplicate_wildcards_does_not_create_error(dataset: BidsDataset, tmpdir: Path):
    root = tempfile.mkdtemp(dir=tmpdir)
    rooted = BidsDataset.from_iterable(
        attrs.evolve(comp, path=os.path.join(root, comp.path))
        for comp in dataset.values()
    )
    create_dataset(Path("/"), rooted)
    config = create_snakebids_config(dataset)
    wildcards = itx.first(config.values()).get("wildcards", [])
    wildcards.append(wildcards[0])
    reindexed = generate_inputs(
        root,
        config,
    )
    assert reindexed == rooted
    assert reindexed.layout is not None


class TestAbsentConfigEntries:
    def get_entities(self, root: Path):
        # Generate directory
        entities = {"subject": ["001", "002"], "acq": sorted(["foo", "bar"])}
        zip_list: dict[str, list[str]] = defaultdict(list)
        for e in it.product(*entities.values()):
            d = dict(zip(entities.keys(), e, strict=True))
            for key, val in d.items():
                zip_list[key].append(val)
            path = Path(bids(root, datatype="anat", suffix="T1w.nii.gz", **d))
            path.parent.mkdir(parents=True, exist_ok=True)
            path.touch()
        return entities, MultiSelectDict(zip_list)

    def test_missing_filters(self, tmpdir: Path):
        _, zip_list = self.get_entities(tmpdir)

        # create config
        derivatives = False
        pybids_inputs: InputsConfig = {
            "t1": {
                "wildcards": ["acquisition", "subject"],
            }
        }

        # Simplest case -- one input type, using pybids
        config = generate_inputs(
            pybids_inputs=pybids_inputs,
            bids_dir=tmpdir,
            derivatives=derivatives,
            pybids_config=str(Path(__file__).parent / "data" / "custom_config.json"),
        )
        template = BidsDataset(
            {"t1": BidsComponent(name="t1", path=config["t1"].path, zip_lists=zip_list)}
        )
        # Order of the subjects is not deterministic
        assert template == config
        assert config.subj_wildcards == {"subject": "{subject}"}

    def test_missing_wildcards(self, tmpdir: Path):
        self.get_entities(tmpdir)

        # create config
        derivatives = False
        pybids_inputs: InputsConfig = {
            "t1": {
                "filters": {"acquisition": "foo", "subject": "001"},
            }
        }

        # Simplest case -- one input type, using pybids
        config = generate_inputs(
            bids_dir=tmpdir,
            pybids_inputs=pybids_inputs,
            derivatives=derivatives,
            pybids_config=str(Path(__file__).parent / "data" / "custom_config.json"),
        )
        template = BidsDataset(
            {"t1": BidsComponent(name="t1", path=config["t1"].path, zip_lists={})}
        )
        assert template == config
        assert config.subj_wildcards == {"subject": "{subject}"}


class TestPostfilter:
    valid_chars = st.characters(blacklist_characters=["\n"])
    st_lists_or_text = st.lists(st.text(valid_chars)) | st.text(valid_chars)

    @given(st.text(), st_lists_or_text)
    def test_returns_participant_label_as_dict(self, key: str, label: list[str] | str):
        filters = PostFilter()
        filters.add_filter(key, label, None)
        if isinstance(label, str):
            assert filters.inclusions == {key: [label]}
        else:
            assert filters.inclusions == {key: label}
        assert filters.exclusions == {}

    @given(
        st.text(),
        st_lists_or_text,
        st.lists(st.text(valid_chars, min_size=1), min_size=1),
        st.text(valid_chars, min_size=1, max_size=3),
    )
    def test_exclude_gives_regex_that_matches_anything_except_exclude(
        self, key: str, excluded: list[str] | str, dummy_values: list[str], padding: str
    ):
        filters = PostFilter()
        # Make sure the dummy_values and padding we'll be testing against are different
        # from our test values
        for value in dummy_values:
            assume(value not in itx.always_iterable(excluded))
        assume(padding not in itx.always_iterable(excluded))

        filters.add_filter(key, None, excluded)
        assert isinstance(filters.exclusions[key], list)
        assert len(filters.exclusions[key]) == 1

        # We match any value that isn't the exclude string
        for value in dummy_values:
            assert re.match(filters.exclusions[key][0], value)

        for exclude in itx.always_iterable(excluded):
            # We don't match the exclude string
            assert re.match(filters.exclusions[key][0], exclude) is None

            # Addition of random strings before and/or after lets the match occur again
            assert re.match(filters.exclusions[key][0], padding + exclude)
            assert re.match(filters.exclusions[key][0], exclude + padding)
            assert re.match(filters.exclusions[key][0], padding + exclude + padding)
        assert filters.inclusions == {}


class PathEntities(NamedTuple):
    entities: dict[str, list[str]]
    template: Path
    filters: dict[str, list[str]]


@st.composite
def path_entities(draw: st.DrawFn):
    """Generate path wildcard entities and corresponding values

    This has three main outputs. The first is a dict of wildcards->values, where values
    are sample values of wildcards. Wildcards are substituteable, {brace-enclosed} str
    for use in string.format.

    The wildcards in the entity dict produced above are combined to form a path name:

    wildcard1-{wildcard1}_wildcard2-{wildcard2}_...

    In addition to this name, a random subsample of wildcards will be selected as dir
    entities. These, along with the path name, will be combined to return a path
    template:

    {wildcard2}/{wildcard5}/wildcard1-{wildcard1}_wildcard2-{wildcard2}_...

    Finally, a subsample of wildcards, along with a subset of their respective values,
    will be selected as filters and returned as a dict of wildcard->values.

    Parameters
    ----------
    draw : st.DrawFn
        [description]

    Returns
    -------
    [type]
        [description]
    """
    # TODO: Remove restrictions on characters to accurately reflect what the function
    #       should allow
    valid_chars = st.characters(
        min_codepoint=48, max_codepoint=122, whitelist_categories=["Ll", "Lu"]
    )
    # We need to explicitly exclude keywords here because the current implementation
    # of glob_wildcards uses a named tuple, which doesn't allow keywords as attributes.
    path_text = st.text(valid_chars, min_size=1).filter(
        lambda s: not keyword.iskeyword(s)
    )
    entities = draw(
        st.dictionaries(
            path_text,
            st.lists(path_text, min_size=1, max_size=2, unique=True),
            min_size=1,
            max_size=5,
        )
    )

    def get_subset(of: Iterable[T]) -> list[T]:
        return draw(
            st.lists(st.sampled_from([*of]), unique=True, max_size=len(entities))
        )

    # Collect dir_entities and filters
    dir_entities = get_subset(entities.keys())
    filtered_entities = get_subset(entities.keys())
    filter_selections = [get_subset(entities[entity]) for entity in filtered_entities]
    filters = dict(zip(filtered_entities, filter_selections, strict=True))

    # Compose the path template
    dir_template = Path(*(f"{{{entity}}}" for entity in dir_entities))
    name_template = "_".join(f"{entity}-{{{entity}}}" for entity in entities)
    template = dir_template / name_template

    return PathEntities(entities, template, filters)


class TestCustomPaths:
    @pytest.fixture
    def temp_dir(self, fakefs_tmpdir: Path, bids_fs: Path):
        return fakefs_tmpdir

    def generate_test_directory(
        self, entities: dict[str, list[str]], template: Path, tmpdir: Path
    ):
        root = Path(tempfile.mkdtemp(prefix="hypothesis-", dir=tmpdir))
        # Generate fake directory structure
        for values in it.product(*entities.values()):
            name_value = dict(zip(entities.keys(), values, strict=True))
            path = str(template).format(**name_value)
            (root / path).parent.mkdir(parents=True, exist_ok=True)
            (root / path).touch()
        return root / template

    T = TypeVar("T")

    def test_benchmark_test_custom_paths(self, benchmark: Benchmark, tmp_path: Path):
        entities = {"A": ["A", "B", "C"], "B": ["1", "2", "3"]}
        template = Path("{A}/A-{A}_B-{B}")
        test_path = self.generate_test_directory(entities, template, tmp_path)
        benchmark(_parse_custom_path, test_path, UnifiedFilter.from_filter_dict({}))

    @allow_function_scoped
    @given(path_entities=path_entities())
    def test_collects_all_paths_when_no_filters(
        self,
        path_entities: PathEntities,
        temp_dir: Path,
    ):
        entities, template, _ = path_entities
        test_path = self.generate_test_directory(entities, template, temp_dir)

        # Test without any filters
        result = _parse_custom_path(test_path, UnifiedFilter.from_filter_dict({}))
        zip_lists = get_zip_list(entities, it.product(*entities.values()))
        assert BidsComponent(
            name="foo", path=get_bids_path(zip_lists), zip_lists=zip_lists
        ) == BidsComponent(
            name="foo", path=get_bids_path(result), zip_lists=MultiSelectDict(result)
        )

    @settings(deadline=400, suppress_health_check=[HealthCheck.function_scoped_fixture])
    @given(path_entities=path_entities())
    def test_collects_only_filtered_entities(
        self,
        path_entities: PathEntities,
        temp_dir: Path,
    ):
        entities, template, filters = path_entities
        test_path = self.generate_test_directory(entities, template, temp_dir)

        # Test with filters
        result_filtered = MultiSelectDict(
            _parse_custom_path(test_path, UnifiedFilter.from_filter_dict(filters))
        )
        zip_lists = MultiSelectDict(
            # Start with empty lists for each key, otherwise keys will be missing
            {key: list[str]() for key in entities}
            |
            # Override entities with relevant filters before making zip lists
            get_zip_list(entities, it.product(*(entities | filters).values()))
        )
        assert BidsComponent(
            name="foo", path=get_bids_path(zip_lists), zip_lists=zip_lists
        ) == BidsComponent(
            name="foo", path=get_bids_path(result_filtered), zip_lists=result_filtered
        )

    @settings(
        deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture]
    )
    @given(path_entities=path_entities())
    def test_collect_all_but_filters_when_exclusion_filters_used(
        self,
        path_entities: PathEntities,
        temp_dir: Path,
    ):
        entities, template, filters = path_entities
        test_path = self.generate_test_directory(entities, template, temp_dir)
        # Test with exclusion filters
        exclude_filters = PostFilter()
        for key, values in filters.items():
            exclude_filters.add_filter(key, None, values)
        result_excluded = MultiSelectDict(
            _parse_custom_path(
                test_path, UnifiedFilter.from_filter_dict({}, exclude_filters)
            )
        )

        entities_excluded = {
            entity: [value for value in values if value not in filters.get(entity, [])]
            for entity, values in entities.items()
        }
        zip_lists = MultiSelectDict(
            # Start with empty lists for each key, otherwise keys will be missing
            {key: list[str]() for key in entities}
            |
            # Override entities with relevant filters before making zip lists
            get_zip_list(entities, it.product(*entities_excluded.values())),
        )

        assert BidsComponent(
            name="foo", path=get_bids_path(zip_lists), zip_lists=zip_lists
        ) == BidsComponent(
            name="foo", path=get_bids_path(result_excluded), zip_lists=result_excluded
        )

    @given(
        boolean=st.booleans(),
        filter=st.none() | st.lists(st.text()),
        path_entities=path_entities(),
    )
    @allow_function_scoped
    def test_errors_when_bools_given_as_filters(
        self,
        temp_dir: Path,
        path_entities: PathEntities,
        boolean: bool,
        filter: list[str] | None,
    ):
        entities, template, _ = path_entities
        test_path = self.generate_test_directory(entities, template, temp_dir)
        with pytest.raises(ValueError, match="Boolean filters in items with custom "):
            _parse_custom_path(
                test_path,
                UnifiedFilter.from_filter_dict(
                    {"foo": boolean if filter is None else [*filter, boolean]}
                ),
            )


def test_custom_pybids_config(tmpdir: Path):
    # Generate directory
    for i in range(2):
        path = Path(
            bids(
                tmpdir, datatype="anat", subject="001", foo=str(i), suffix="T1w.nii.gz"
            )
        )
        path.parent.mkdir(parents=True, exist_ok=True)
        path.touch()

    # create config
    derivatives = False
    pybids_inputs: InputsConfig = {
        "t1": {
            "filters": {"suffix": "T1w"},
            "wildcards": ["acquisition", "subject", "foo"],
        }
    }

    # Simplest case -- one input type, using pybids
    result = generate_inputs(
        pybids_inputs=pybids_inputs,
        bids_dir=tmpdir,
        derivatives=derivatives,
        pybids_config=str(Path(__file__).parent / "data" / "custom_config.json"),
    )
    template = BidsDataset(
        {
            "t1": BidsComponent(
                name="t1",
                path=bids(
                    tmpdir,
                    datatype="anat",
                    subject="{subject}",
                    foo="{foo}",
                    suffix="T1w.nii.gz",
                ),
                zip_lists={"foo": ["0", "1"], "subject": ["001", "001"]},
            )
        }
    )
    assert template == result
    assert result["t1"].wildcards == {"foo": "{foo}", "subject": "{subject}"}
    # Order of the subjects is not deterministic
    assert result.subj_wildcards == {"subject": "{subject}"}


def test_nonstandard_custom_pybids_config(tmpdir: Path):
    # Generate directory
    for i in range(2):
        path = Path(
            bids(
                tmpdir, datatype="anat", subject="001", foo=str(i), suffix="T1w.nii.gz"
            )
        )
        path.parent.mkdir(parents=True, exist_ok=True)
        path.touch()

    # create config
    derivatives = False
    pybids_inputs: InputsConfig = {
        "t1": {
            "filters": {"suffix": "T1w"},
            "wildcards": ["acquisition", "subject", "foobar"],
        }
    }

    # Simplest case -- one input type, using pybids
    with pytest.raises(ConfigError):
        generate_inputs(
            pybids_inputs=pybids_inputs,
            bids_dir=tmpdir,
            derivatives=derivatives,
            pybids_config=(
                str(Path(__file__).parent / "data" / "custom_config_nonstandard.json")
            ),
        )


def test_index_metadata(mocker: MockerFixture):
    spy = mocker.spy(input_generation, "BIDSLayoutIndexer")
    mocker.patch.object(input_generation, "BIDSLayout", side_effect=ValueError)

    # Simplest case -- one input type, using pybids
    with pytest.raises(ValueError):  # noqa
        generate_inputs(
            pybids_inputs={"foo": {}},
            bids_dir=...,  # type: ignore
            derivatives=...,  # type: ignore
            index_metadata=True,
        )

    spy.assert_called_once_with(
        validate=False,
        index_metadata=True,
    )


def test_t1w():
    # create config
    real_bids_dir = "tests/data/bids_t1w"
    derivatives = False
    pybids_inputs: InputsConfig = {
        "t1": {
            "filters": {"suffix": "T1w"},
            "wildcards": ["acquisition", "subject", "session", "run"],
        }
    }

    # Simplest case -- one input type, using pybids
    result = generate_inputs(
        pybids_inputs=pybids_inputs,
        bids_dir=real_bids_dir,
        derivatives=derivatives,
    )
    template = BidsDataset(
        {
            "t1": BidsComponent(
                name="t1",
                path=result["t1"].path,
                zip_lists={"acq": ["mprage", "mprage"], "subject": ["001", "002"]},
            )
        }
    )
    assert template == result

    # Order of the subjects is not deterministic
    assert result.subjects in [["001", "002"], ["002", "001"]]
    assert result.sessions == []
    assert result.subj_wildcards == {"subject": "{subject}"}

    pybids_inputs_suffix: InputsConfig = {
        "scan": {
            "filters": {},
            "wildcards": [
                "acquisition",
                "subject",
                "session",
                "run",
                "suffix",
            ],
        }
    }
    result = generate_inputs(
        pybids_inputs=pybids_inputs_suffix,
        bids_dir=real_bids_dir,
        derivatives=derivatives,
        participant_label="001",
    )
    assert result["scan"].entities == {
        "acq": ["mprage"],
        "subject": ["001"],
        "suffix": ["T1w"],
    }
    template = BidsDataset(
        {
            "scan": BidsComponent(
                name="scan",
                path=result["scan"].path,
                zip_lists={
                    "acq": [
                        "mprage",
                    ],
                    "subject": [
                        "001",
                    ],
                    "suffix": [
                        "T1w",
                    ],
                },
            )
        }
    )
    assert template == result

    assert result.subjects == ["001"]
    assert result.sessions == []
    assert result.subj_wildcards == {"subject": "{subject}"}

    # Two input types, specified by pybids or path override
    wildcard_path_t1 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "data/bids_t1w",
        "sub-{subject}/anat/sub-{subject}_acq-{acq}_T1w.nii.gz",
    )
    wildcard_path_t2 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "data/bids_t1w",
        "sub-{subject}/anat/sub-{subject}_T2w.nii.gz",
    )
    pybids_inputs = {
        "t1": {
            "filters": {"suffix": "T1w"},
            "wildcards": ["acquisition", "subject", "session", "run"],
        },
        "t2": {
            "filters": {"suffix": "T2w"},
            "wildcards": ["acquisition", "subject", "session", "run"],
        },
    }
    bids_dir = real_bids_dir

    # Want to test both inputs from layout, both inputs from custom path, and
    # one of each. This setup should produce the same results every time.
    for idx in range(4):
        if idx == 1:
            pybids_inputs["t1"]["custom_path"] = wildcard_path_t1
        elif idx == 2:
            pybids_inputs["t2"]["custom_path"] = wildcard_path_t2
        elif idx == 3:
            pybids_inputs["t1"]["custom_path"] = wildcard_path_t1
            pybids_inputs["t2"]["custom_path"] = wildcard_path_t2
            # TODO: Allow arbitrary paths to work when all custom paths are used
            # bids_dir = "-"
        result = generate_inputs(
            pybids_inputs=pybids_inputs,
            bids_dir=bids_dir,
            derivatives=derivatives,
        )
        template = BidsDataset(
            {
                "t1": BidsComponent(
                    name="t1",
                    path=result["t1"].path,
                    zip_lists={
                        "acq": ["mprage", "mprage"],
                        "subject": ["001", "002"],
                    },
                ),
                "t2": BidsComponent(
                    name="t2",
                    path=result["t2"].path,
                    zip_lists={"subject": ["002"]},
                ),
            }
        )
        # Order of the subjects is not deterministic
        assert result.subjects in [["001", "002"], ["002", "001"]]

        assert result.sessions == []
        assert result.subj_wildcards == {"subject": "{subject}"}


def test_t1w_with_dict():
    # create config
    real_bids_dir = "tests/data/bids_t1w"
    derivatives = False
    pybids_inputs: InputsConfig = {
        "t1": {
            "filters": {"suffix": "T1w"},
            "wildcards": ["acquisition", "subject", "session", "run"],
        }
    }

    # Simplest case -- one input type, using pybids
    config = generate_inputs(
        pybids_inputs=pybids_inputs,
        bids_dir=real_bids_dir,
        derivatives=derivatives,
        use_bids_inputs=False,
    )
    # Order of the subjects is not deterministic
    assert config["input_lists"] == BidsListCompare(
        {"t1": {"acq": ["mprage"], "subject": ["002", "001"]}}
    )
    assert config["input_zip_lists"] == {
        "t1": {"acq": ["mprage", "mprage"], "subject": ["001", "002"]}
    }
    assert config["input_wildcards"] == {"t1": {"acq": "{acq}", "subject": "{subject}"}}
    # Order of the subjects is not deterministic
    assert set(config["subjects"]) == {"002", "001"}
    assert config["sessions"] == []
    assert config["subj_wildcards"] == {"subject": "{subject}"}

    pybids_inputs_suffix: InputsConfig = {
        "scan": {
            "filters": {},
            "wildcards": [
                "acquisition",
                "subject",
                "session",
                "run",
                "suffix",
            ],
        }
    }
    config = generate_inputs(
        pybids_inputs=pybids_inputs_suffix,
        bids_dir=real_bids_dir,
        derivatives=derivatives,
        participant_label="001",
        use_bids_inputs=False,
    )
    assert config["input_lists"] == {
        "scan": {"acq": ["mprage"], "subject": ["001"], "suffix": ["T1w"]}
    }
    assert config["input_zip_lists"] == {
        "scan": {"acq": ["mprage"], "subject": ["001"], "suffix": ["T1w"]}
    }
    assert config["input_wildcards"] == {
        "scan": {"acq": "{acq}", "subject": "{subject}", "suffix": "{suffix}"}
    }
    assert config["subjects"] == ["001"]
    assert config["sessions"] == []
    assert config["subj_wildcards"] == {"subject": "{subject}"}


def test_get_lists_from_bids_raises_run_error():
    bids_layout = None
    pybids_inputs: InputsConfig = {
        "t1": {
            "filters": {"suffix": "T1w"},
            "wildcards": ["acquisition", "subject", "session", "run"],
        }
    }
    with pytest.raises(RunError):
        next(
            _get_components(
                bids_layout=bids_layout,
                inputs_config=pybids_inputs,
                postfilters=PostFilter(),
            )
        )


def test_get_lists_from_bids():
    bids_dir = "tests/data/bids_t1w"
    wildcard_path_t1 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "data/bids_t1w",
        "sub-{subject}/anat/sub-{subject}_acq-{acq}_T1w.nii.gz",
    )
    wildcard_path_t2 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "data/bids_t1w",
        "sub-{subject}/anat/sub-{subject}_T2w.nii.gz",
    )
    print(wildcard_path_t1)
    layout = BIDSLayout(bids_dir, validate=False)
    pybids_inputs: InputsConfig = {
        "t1": {
            "filters": {"suffix": "T1w"},
            "wildcards": ["acquisition", "subject", "session", "run"],
        },
        "t2": {
            "filters": {"suffix": "T2w"},
            "wildcards": ["acquisition", "subject", "session", "run"],
        },
    }

    # Want to test both inputs from layout, both inputs from custom path, and
    # one of each. This setup should produce the same results every time.
    for idx in range(4):
        if idx == 1:
            pybids_inputs["t1"]["custom_path"] = wildcard_path_t1
        elif idx == 2:
            pybids_inputs["t2"]["custom_path"] = wildcard_path_t2
        elif idx == 3:
            pybids_inputs["t1"]["custom_path"] = wildcard_path_t1
            pybids_inputs["t2"]["custom_path"] = wildcard_path_t2

        result = _get_components(
            bids_layout=layout, inputs_config=pybids_inputs, postfilters=PostFilter()
        )
        for bids_lists in result:
            if bids_lists.input_name == "t1":
                template = BidsComponent(
                    name="t1",
                    path=wildcard_path_t1,
                    zip_lists={
                        "acq": ["mprage", "mprage"],
                        "subject": ["001", "002"],
                    },
                )
                assert template == bids_lists
            elif bids_lists.input_name == "t2":
                assert bids_lists.input_path == wildcard_path_t2
                template = BidsComponent(
                    name="t2",
                    path=wildcard_path_t2,
                    zip_lists={
                        "subject": ["002"],
                    },
                )
                assert template == bids_lists


class TestGenBidsLayout:
    @settings(
        max_examples=1, suppress_health_check=[HealthCheck.function_scoped_fixture]
    )
    @given(dataset=sb_st.datasets())
    def test_gen_layout_returns_valid_dataset(self, dataset: BidsDataset, tmpdir: Path):
        create_dataset(tmpdir, dataset)
        assert _gen_bids_layout(
            bids_dir=tmpdir,
            derivatives=False,
            pybidsdb_dir=None,
            pybidsdb_reset=False,
            pybids_config=None,
        )

    def test_invalid_path_raises_error(self, tmpdir: Path):
        with pytest.raises(ValueError, match="BIDS root does not exist"):
            _gen_bids_layout(
                bids_dir=tmpdir / "foo",
                derivatives=False,
                pybidsdb_dir=None,
                pybidsdb_reset=False,
            )


@pytest.mark.parametrize("count", tuple(range(6)))
def test_all_custom_paths(count: int):
    config: InputsConfig = {}
    for i in range(count):
        config[str(i)] = {"wildcards": [], "filters": {}, "custom_path": "foo"}
    for i in range(count, 5):
        config[str(i)] = {
            "wildcards": [],
            "filters": {},
        }
    if count == 5:
        assert _all_custom_paths(config)
    else:
        assert not _all_custom_paths(config)


class TestRecogPathSchemes:
    PATH_AND_TYPES = (
        ("file", "RELATIVE"),
        ("hello", "RELATIVE"),
        ("gs", "RELATIVE"),
        ("./hello/world", "RELATIVE"),
        ("hello/world", "RELATIVE"),
        ("/hello/world", "ABSOLUTE"),
        ("gs://some/google/cloud/bucket", "NETWORK"),
        ("s3://some/s3/bucket", "NETWORK"),
    )

    @pytest.mark.parametrize(("path", "path_type"), PATH_AND_TYPES)
    def test_is_local_relative(self, path: str, path_type: str):
        isnet = path_type == "NETWORK"
        is_local_relative = path_type == "RELATIVE"

        # test the path itself, and the corresponding Path(path)
        assert _is_local_relative(path) == is_local_relative, (
            f"Path {path} fails is local relative path test."
        )
        if not isnet:
            assert _is_local_relative(Path(path)) == is_local_relative

    @pytest.mark.skipif(
        sys.version_info < (3, 12), reason="Path class has no with_segments()"
    )
    @pytest.mark.parametrize(
        ("path", "path_type"), [tup for tup in PATH_AND_TYPES if tup[1] == "RELATIVE"]
    )
    def test_path_subclassing(self, path: str, path_type: str):
        # Google cloud is not posix, for mocking purpose however we just
        # need a class that is a subclass of Path
        class MockGCSPath(PosixPath):
            def __init__(self, *pathsegments: str):  # pyright: ignore[reportInconsistentConstructor]
                super().__init__(*pathsegments)

            def __str__(self):  # __fspath__ calls __str__ by default
                return f"gs://{super().__str__()}"

        assert not _is_local_relative(MockGCSPath(path))


@example(
    dataset=BidsDataset(
        {
            "1": BidsComponent(
                name="1",
                path="sub-{subject}/sub-{subject}_{suffix}{extension}",
                zip_lists={
                    "subject": ["0"],
                    "suffix": ["0"],
                    "extension": [".0"],
                },
            ),
            "0": BidsComponent(
                name="0",
                path="sub-{subject}/sub-{subject}{extension}",
                zip_lists={
                    "subject": ["0"],
                    "extension": [".0"],
                },
            ),
        }
    ),
)
@settings(
    deadline=800,
    suppress_health_check=[
        HealthCheck.function_scoped_fixture,
        HealthCheck.too_slow,
    ],
)
@given(dataset=sb_st.datasets(unique=True))
def test_generate_inputs(dataset: BidsDataset, tmpdir: Path):
    root = tempfile.mkdtemp(dir=tmpdir)
    rooted = BidsDataset.from_iterable(
        attrs.evolve(comp, path=os.path.join(root, comp.path))
        for comp in dataset.values()
    )
    reindexed = reindex_dataset(root, rooted)
    assert reindexed == rooted
    assert reindexed.layout is not None


@st.composite
def dataset_with_subject(draw: st.DrawFn):
    entities = draw(sb_st.bids_entity_lists(blacklist_entities=["subject"]))
    entities += ["subject"]
    return BidsDataset.from_iterable(
        [
            draw(
                sb_st.bids_components(
                    whitelist_entities=entities,
                    min_entities=len(entities),
                    max_entities=len(entities),
                    restrict_patterns=True,
                    unique=True,
                )
            )
        ]
    )


class TestParticipantFiltering:
    MODE = Literal["include", "exclude"]

    def get_filter_params(self, mode: MODE, filters: list[str] | str):
        class FiltParams(TypedDict, total=False):
            participant_label: list[str] | str
            exclude_participant_label: list[str] | str

        if mode == "include":
            return FiltParams({"participant_label": filters})
        if mode == "exclude":
            return FiltParams({"exclude_participant_label": filters})
        msg = f"Invalid mode specification: {mode}"
        raise ValueError(msg)

    @given(
        data=st.data(),
        dataset=dataset_with_subject(),
    )
    @settings(
        deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture]
    )
    def test_exclude_and_participant_label_filter_correctly(
        self, data: st.DataObject, dataset: BidsDataset, tmpdir: Path
    ):
        root = tempfile.mkdtemp(dir=tmpdir)
        rooted = BidsDataset.from_iterable(
            attrs.evolve(comp, path=os.path.join(root, comp.path))
            for comp in dataset.values()
        )
        sampler = st.sampled_from(itx.first(rooted.values()).entities["subject"])
        excluded = data.draw(st.lists(sampler, unique=True) | sampler | st.none())
        included = data.draw(st.lists(sampler, unique=True) | sampler | st.none())
        reindexed = reindex_dataset(
            root, rooted, exclude_participant_label=excluded, participant_label=included
        )
        reindexed_subjects = set(itx.first(reindexed.values()).entities["subject"])
        expected_subjects = set(itx.first(rooted.values()).entities["subject"])
        if included is not None:
            expected_subjects &= set(itx.always_iterable(included))
        if excluded is not None:
            expected_subjects -= set(itx.always_iterable(excluded))

        assert reindexed_subjects == expected_subjects

    @pytest.mark.parametrize("mode", ["include", "exclude"])
    @given(
        dataset=sb_st.datasets_one_comp(blacklist_entities=["subject"], unique=True),
        participant_filter=st.lists(st.text(min_size=1)) | st.text(min_size=1),
    )
    @settings(
        deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture]
    )
    def test_participant_label_doesnt_filter_comps_without_subject(
        self,
        mode: MODE,
        dataset: BidsDataset,
        participant_filter: list[str] | str,
        tmpdir: Path,
    ):
        root = tempfile.mkdtemp(dir=tmpdir)
        rooted = BidsDataset.from_iterable(
            attrs.evolve(comp, path=os.path.join(root, comp.path))
            for comp in dataset.values()
        )
        reindexed = reindex_dataset(
            root, rooted, **self.get_filter_params(mode, participant_filter)
        )
        assert reindexed == rooted

    @pytest.mark.parametrize("mode", ["include", "exclude"])
    @given(
        dataset=dataset_with_subject(),
        participant_filter=st.lists(st.text(min_size=1)) | st.text(min_size=1),
    )
    @settings(
        deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture]
    )
    def test_participant_label_doesnt_filter_comps_when_subject_has_filter(
        self,
        mode: MODE,
        dataset: BidsDataset,
        participant_filter: list[str] | str,
        tmpdir: Path,
    ):
        root = tempfile.mkdtemp(dir=tmpdir)
        rooted = BidsDataset.from_iterable(
            attrs.evolve(comp, path=os.path.join(root, comp.path))
            for comp in dataset.values()
        )
        create_dataset(Path("/"), rooted)
        reindexed = generate_inputs(
            root,
            create_snakebids_config(rooted),
            **self.get_filter_params(mode, participant_filter),
        )
        assert reindexed == rooted

    @pytest.mark.parametrize("mode", ["include", "exclude"])
    @given(
        dataset=dataset_with_subject(),
        participant_filter=st.lists(st.text(min_size=1)) | st.text(min_size=1),
        data=st.data(),
    )
    @settings(
        deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture]
    )
    def test_participant_label_doesnt_filter_comps_when_subject_has_filter_no_wcard(
        self,
        mode: MODE,
        dataset: BidsDataset,
        participant_filter: list[str] | str,
        data: st.DataObject,
        tmpdir: Path,
    ):
        root = tempfile.mkdtemp(dir=tmpdir)
        rooted = BidsDataset.from_iterable(
            attrs.evolve(comp, path=os.path.join(root, comp.path))
            for comp in dataset.values()
        )
        subject = data.draw(
            st.sampled_from(itx.first(rooted.values()).entities["subject"])
        )
        create_dataset(Path("/"), rooted)
        config = create_snakebids_config(rooted)
        for comp in config.values():
            comp["filters"] = dict(comp.get("filters", {}))
            comp["filters"]["subject"] = subject
        reindexed = generate_inputs(
            root,
            create_snakebids_config(rooted),
            **self.get_filter_params(mode, participant_filter),
        )
        assert reindexed == rooted

    @given(
        data=st.data(),
        dataset=dataset_with_subject().filter(
            lambda ds: set(itx.first(ds.values()).wildcards) != {"subject", "extension"}
        ),
    )
    @settings(
        deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture]
    )
    def test_exclude_participant_does_not_make_all_other_filters_regex(
        self, data: st.DataObject, dataset: BidsDataset, tmpdir: Path
    ):
        root = tempfile.mkdtemp(dir=tmpdir)
        rooted = BidsDataset.from_iterable(
            attrs.evolve(comp, path=os.path.join(root, comp.path))
            for comp in dataset.values()
        )

        # Create an extra set of paths by modifying one of the existing components to
        # put foo after a set of entity values. If that filter gets changed to a regex,
        # all of the suffixed decoys will get picked up by pybids
        ziplist = dict(itx.first(rooted.values()).zip_lists)
        mut_entity = itx.first(
            filter(lambda e: e not in {"subject", "extension"}, ziplist)
        )
        ziplist[mut_entity] = ["foo" + v for v in ziplist[mut_entity]]
        for path in sb_expand(itx.first(rooted.values()).path, zip, **ziplist):
            p = Path(path)
            p.parent.mkdir(parents=True, exist_ok=True)
            p.touch()

        sampler = st.sampled_from(itx.first(rooted.values()).entities["subject"])
        label = data.draw(st.lists(sampler, unique=True) | sampler)
        reindexed = reindex_dataset(root, rooted, exclude_participant_label=label)
        reindexed_subjects = set(itx.first(reindexed.values()).entities["subject"])
        original_subjects = set(itx.first(rooted.values()).entities["subject"])
        assert reindexed_subjects == original_subjects - set(itx.always_iterable(label))


# The content of the dataset is irrelevant to this test, so one example suffices
# but can't use extension because the custom path won't glob properly
@settings(max_examples=1, suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(dataset=sb_st.datasets_one_comp(blacklist_entities=["extension"], unique=True))
def test_when_all_custom_paths_no_layout_indexed(
    dataset: BidsDataset, tmpdir: Path, mocker: MockerFixture
):
    # Need to reset mocker at beginning because hypothesis may call this function
    # multiple times
    mocker.stopall()
    root = tempfile.mkdtemp(dir=tmpdir)
    rooted = BidsDataset.from_iterable(
        attrs.evolve(comp, path=os.path.join(root, comp.path))
        for comp in dataset.values()
    )

    spy = mocker.spy(BIDSLayout, "__init__")
    reindexed = reindex_dataset(root, rooted, use_custom_paths=True)
    assert reindexed == rooted
    assert reindexed.layout is None
    spy.assert_not_called()


class TestParseBidsPath:
    @given(
        component=sb_st.bids_components(max_values=1, restrict_patterns=True),
        scheme=sb_st.schemes() | st.none(),
    )
    def test_splits_wildcards_from_path(
        self, component: BidsComponent, scheme: str | None
    ):
        path = component.expand()[0]
        if scheme is not None:
            path = f"{scheme}{path}"
        entities = [BidsEntity.normalize(e).entity for e in component.zip_lists]
        tpl_path, matches = _parse_bids_path(path, entities)
        assert tpl_path.format(**matches) == path

    @given(
        component=sb_st.bids_components(max_values=1, restrict_patterns=True),
        scheme=sb_st.schemes() | st.none(),
    )
    def test_one_match_found_for_each_entity(
        self, component: BidsComponent, scheme: str | None
    ):
        path = component.expand()[0]
        if scheme is not None:
            path = f"{scheme}{path}"
        entities = [BidsEntity.normalize(e).entity for e in component.zip_lists]
        _, matches = _parse_bids_path(path, entities)
        assert set(matches.items()) == {
            (key, val[0]) for key, val in component.zip_lists.items()
        }

    @given(
        component=sb_st.bids_components(
            max_values=1, restrict_patterns=True, extra_entities=False
        ),
        scheme=sb_st.schemes() | st.none(),
        entity=sb_st.bids_entity(),
    )
    def test_missing_match_leads_to_error(
        self, component: BidsComponent, scheme: str | None, entity: BidsEntity
    ):
        path = component.expand()[0]
        if scheme is not None:
            path = f"{scheme}{path}"
        entities = [BidsEntity.normalize(e).entity for e in component.zip_lists]
        assume(entity.entity not in entities)
        with pytest.raises(BidsParseError) as err:
            _parse_bids_path(path, it.chain(entities, [entity.entity]))
        assert err.value.entity == entity


class TestDB:
    @pytest.fixture(autouse=True)
    def _start(self, tmp_path: Path):
        self.tmpdir = str(tmp_path)

        # Copy over test data
        shutil.copytree("tests/data/bids_t1w", f"{self.tmpdir}/data")
        assert filecmp.dircmp("tests/data/bids_t1w", f"{self.tmpdir}/data")

        # Create config
        self.bids_dir: str = f"{self.tmpdir}/data"
        self.pybidsdb_dir = ""
        self.pybidsdb_reset = False

    def test_pybidsdb_dir_blank(self):
        # Test non-saving (check db does not exist)
        _gen_bids_layout(
            bids_dir=self.bids_dir,
            derivatives=False,
            pybidsdb_dir=self.pybidsdb_dir,
            pybidsdb_reset=self.pybidsdb_reset,
        )
        assert not os.path.exists(self.pybidsdb_dir)

    def test_pybidsdb_dir_relative(self):
        # Update config
        self.pybidsdb_dir = "./.db"

        # Check to make sure db exists (relative path)
        _gen_bids_layout(
            bids_dir=self.bids_dir,
            derivatives=False,
            pybidsdb_dir=self.pybidsdb_dir,
            pybidsdb_reset=self.pybidsdb_reset,
        )
        assert not os.path.exists(f"{self.tmpdir}/data/.db/")

    def test_pybidsdb_dir_absolute(self):
        # Update config
        self.pybidsdb_dir = f"{self.tmpdir}/data/.db/"
        self.pybidsdb_reset = False

        # Check to make sure db exists (absolute path)
        _gen_bids_layout(
            bids_dir=self.bids_dir,
            derivatives=False,
            pybidsdb_dir=self.pybidsdb_dir,
            pybidsdb_reset=self.pybidsdb_reset,
        )
        assert os.path.exists(f"{self.tmpdir}/data/.db/")

        # Test reading of old layout when changes occur
        os.makedirs(f"{self.tmpdir}/data/sub-003/anat")
        shutil.copy(
            f"{self.bids_dir}/sub-001/anat/sub-001_acq-mprage_T1w.nii.gz",
            f"{self.bids_dir}/sub-003/anat/sub-003_acq-mprage_T1w.nii.gz",
        )
        # Check to make sure new subject not cached in layout
        layout = _gen_bids_layout(
            bids_dir=self.bids_dir,
            derivatives=False,
            pybidsdb_dir=self.pybidsdb_dir,
            pybidsdb_reset=self.pybidsdb_reset,
        )
        assert not layout.get(subject="003")

        # Test updating of layout
        self.pybidsdb_reset = True
        # Check to see if new subject in updated layout
        layout = _gen_bids_layout(
            bids_dir=self.bids_dir,
            derivatives=False,
            pybidsdb_dir=self.pybidsdb_dir,
            pybidsdb_reset=self.pybidsdb_reset,
        )
        assert layout.get(subject="003")



"""
This type stub file was generated by pyright.
"""

if False: ...

class Version:
    def __init__(self, major: int, minor: int) -> None: ...
    @property
    def major(self) -> int: ...
    @property
    def minor(self) -> int: ...
    def __eq__(self, v: Any) -> bool: ...
    def __lt__(self, v: Version) -> bool: ...
    def __le__(self, v: Version) -> bool: ...
    def __gt__(self, v: Version) -> bool: ...
    def __ge__(self, v: Version) -> bool: ...

def version(
    major: int | str | Tuple[int, int] | None, minor: Optional[int] = ...
) -> Optional[Version]: ...

class Tag:
    def __init__(self, handle: str, prefix: str) -> None: ...
    @property
    def handle(self) -> str: ...
    @property
    def prefix(self) -> str: ...

class DocInfo:
    """
    Store document information, can be used for analysis of a loaded YAML document
    requested_version: if explicitly set before load
    doc_version: from %YAML directive
    tags: from %TAG directives in scanned order
    """
    def __init__(
        self,
        requested_version: Optional[Version] = ...,
        doc_version: Optional[Version] = ...,
        tags: Optional[list[Tag]] = ...,
    ) -> None: ...



"""
This type stub file was generated by pyright.
"""

""" Utility functions. """

def listify(obj):  # -> list[Unknown] | tuple[Unknown] | None:
    """Wraps all non-list or tuple objects in a list; provides a simple way
    to accept flexible arguments."""
    ...

def matches_entities(obj, entities, strict=...):  # -> bool:
    """Checks whether an object's entities match the input."""
    ...

def natural_sort(l, field=...):  # -> list[Unknown]:
    """
    based on snippet found at http://stackoverflow.com/a/4836734/2445984
    """
    ...

def convert_JSON(j):  # -> dict[Unknown, Unknown]:
    """Recursively convert CamelCase keys to snake_case.
    From: https://stackoverflow.com/questions/17156078/
    converting-identifier-naming-between-camelcase-and-
    underscores-during-json-seria
    """
    ...

def splitext(path):  # -> list[Unknown]:
    """splitext for paths with directories that may contain dots.
    From https://stackoverflow.com/questions/5930036/separating-file-extensions-using-python-os-path-module
    """
    ...

def make_bidsfile(filename):  # -> Any:
    """Create a BIDSFile instance of the appropriate class."""
    ...

def collect_associated_files(layout, files, extra_entities=...):  # -> list[Unknown]:
    """Collect and group BIDSFiles with multiple files per acquisition.

    Parameters
    ----------
    layout
    files : list of BIDSFile
    extra_entities

    Returns
    -------
    collected_files : list of list of BIDSFile
    """
    ...

def validate_multiple(val, retval=...):  # -> tuple[Unknown] | None:
    """Any click.Option with the multiple flag will return an empty tuple if not set.

    This helper method converts empty tuples to a desired return value (default: None).
    This helper method selects the first item in single-item tuples.
    """
    ...



# type: ignore
__submodules__ = ["filtering", "input_generation", "datasets"]

__ignore__ = ["T_co"]


# <AUTOGEN_INIT>
import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

__all__ = [
    "BidsComponent",
    "BidsComponentRow",
    "BidsDataset",
    "BidsDatasetDict",
    "BidsPartialComponent",
    "filter_list",
    "generate_inputs",
    "get_filtered_ziplist_index",
    "get_wildcard_constraints",
    "write_derivative_json",
]
# </AUTOGEN_INIT>



"""
This type stub file was generated by pyright.
"""

if False: ...
__all__ = [
    "FileMark",
    "StringMark",
    "CommentMark",
    "YAMLError",
    "MarkedYAMLError",
    "ReusedAnchorWarning",
    "UnsafeLoaderWarning",
    "MarkedYAMLWarning",
    "MarkedYAMLFutureWarning",
]

class StreamMark:
    __slots__ = ...
    def __init__(self, name: Any, index: int, line: int, column: int) -> None: ...
    def __str__(self) -> Any: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...

class FileMark(StreamMark):
    __slots__ = ...

class StringMark(StreamMark):
    __slots__ = ...
    def __init__(
        self, name: Any, index: int, line: int, column: int, buffer: Any, pointer: Any
    ) -> None: ...
    def get_snippet(self, indent: int = ..., max_length: int = ...) -> Any: ...
    def __str__(self) -> Any: ...
    def __repr__(self) -> Any: ...

class CommentMark:
    __slots__ = ...
    def __init__(self, column: Any) -> None: ...

class YAMLError(Exception): ...

class MarkedYAMLError(YAMLError):
    def __init__(
        self,
        context: Any = ...,
        context_mark: Any = ...,
        problem: Any = ...,
        problem_mark: Any = ...,
        note: Any = ...,
        warn: Any = ...,
    ) -> None: ...
    def __str__(self) -> Any: ...
    def check_append(self, lines: list[str], val: Optional[str]) -> None: ...

class YAMLStreamError(Exception): ...
class YAMLWarning(Warning): ...

class MarkedYAMLWarning(YAMLWarning):
    def __init__(
        self,
        context: Any = ...,
        context_mark: Any = ...,
        problem: Any = ...,
        problem_mark: Any = ...,
        note: Any = ...,
        warn: Any = ...,
    ) -> None: ...
    def __str__(self) -> Any: ...
    def check_append(self, lines: list[str], val: Optional[str]) -> None: ...

class ReusedAnchorWarning(YAMLWarning): ...

class UnsafeLoaderWarning(YAMLWarning):
    text = ...

class MantissaNoDotYAML1_1Warning(YAMLWarning):
    def __init__(self, node: Any, flt_str: Any) -> None: ...
    def __str__(self) -> Any: ...

class YAMLFutureWarning(Warning): ...

class MarkedYAMLFutureWarning(YAMLFutureWarning):
    def __init__(
        self,
        context: Any = ...,
        context_mark: Any = ...,
        problem: Any = ...,
        problem_mark: Any = ...,
        note: Any = ...,
        warn: Any = ...,
    ) -> None: ...
    def __str__(self) -> Any: ...
    def check_append(self, lines: list[str], val: Optional[str]) -> None: ...



from __future__ import annotations

import os
import tempfile
from pathlib import Path

import bids.layout
import pytest
from hypothesis import database, settings
from pyfakefs.fake_filesystem import FakeFilesystem

import snakebids.paths.resources as specs
from snakebids import resources, set_bids_spec

## Hypothesis profiles

# TODO: The default Directory Database would be much better, but it's currently
#       incompatible with pyfakefs. To fix it, we need to either patch @given or add
#       another decorator that only activates the fake filesystem within the test body

# github-actions tends to have flaky runtimes, likely due to temporary slowdowns in the
# runner, so just disable deadlines
settings.register_profile(
    "pr", deadline=None, database=database.InMemoryExampleDatabase()
)
settings.register_profile("dev", database=database.InMemoryExampleDatabase())

settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", "dev"))

set_bids_spec("v0_0_0")


# Fixtures
@pytest.fixture
def fakefs(
    request: pytest.FixtureRequest,
):
    """fakefs wrapper that can be disabled using the mark "disable_fakefs"

    Disabling is useful for debugging, as fakefs messes up the breakpoints in vscode

    Returns
    -------
    FakeFilesystem or None
        None if disabled
    """
    if request.node.get_closest_marker("disable_fakefs"):  # type: ignore
        return None

    return request.getfixturevalue("fs")


@pytest.fixture
def fakefs_tmpdir(
    request: pytest.FixtureRequest, fakefs: FakeFilesystem | None
) -> Path:
    """Version of tmpdir compatible with fakefs

    If fakefs is disabled, a tmpdir is returned using the builtin tmpdir fixture.
    Otherwise, an arbitrary tempdir will be created in the fakefs

    Returns
    -------
    Path
    """
    if not fakefs:
        return Path(request.getfixturevalue("tmpdir"))
    return Path(tempfile.mkdtemp())


@pytest.fixture
def bids_fs(fakefs: FakeFilesystem | None) -> FakeFilesystem | None:
    if fakefs:
        f = Path(*bids.layout.__path__, "config")
        fakefs.add_real_file(f / "bids.json")
        fakefs.add_real_file(f / "derivatives.json")
        fakefs.add_real_file(Path(*resources.__path__) / "bids_tags.json")
        fakefs.add_real_directory(Path(*specs.__path__))
    return fakefs



from argparse import ArgumentParser
from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from types import ModuleType
from typing import Any

from snakemake.common import configfile as configfile  # type: ignore

configfile: ModuleType

class WildcardError(Exception): ...

def load_configfile(configpath: str) -> dict[str, Any]:
    "Load a JSON or YAML configfile as a dict, then checks that it's a dict."

def expand(
    filepatterns: Sequence[Path | str] | Path | str,
    func: Callable[[Iterable[str]], Iterable[Iterable[str]]] | None = ...,
    /,
    *,
    allow_missing: bool | Sequence[str] | str = ...,
    **wildcards: Sequence[str] | str,
) -> list[str]:
    """
    Expand wildcards in given filepatterns.

    Arguments
    *args -- first arg: filepatterns as list or one single filepattern,
        second arg (optional): a function to combine wildcard values
        (itertools.product per default)
    **wildcards -- the wildcards as keyword arguments
        with their values as lists. If allow_missing=True is included
        wildcards in filepattern without values will stay unformatted.
    """

class Namedlist[T](list[T | Iterable[T]]): ...

class InputFiles[T](Namedlist[T]):
    def __getattribute__(self, __name: str) -> str: ...

class OutputFiles[T](Namedlist[T]):
    def __getattribute__(self, __name: str) -> str: ...

class Params[T](Namedlist[T]): ...

class Snakemake:
    input: InputFiles[str]
    output: OutputFiles[str]
    params: Params[str]

def main(argv: list[str] = ...) -> None: ...
def get_argument_parser(profiles: list[str] | None = ...) -> ArgumentParser: ...
